from os import makedirs
import os.path
import hashlib
import json
from liquer.state_types import state_types_registry
from liquer.state import State
from liquer.parser import all_splits, encode, decode
import logging
import traceback
import base64
import numpy as np

"""
Cache defines a mechanism for caching state for a query (or a subquery).
Since there is a one-to-one correspondence between a query and a state (more precisely state data),
cache can work as a simple key-value store, using state.query as a key.
Cache object must define three methods:
- get - for retrieving state from a key (query); get returns None if the state cannot be recovered
- store - for storing the state; should return True if the cache handled the storing 
- contains - for checking the availability of a state associated with a key (query)

A global cache is configured by set_cache and available via get_cache.
From a given query, cached_part function tries to recover as much as possible from cache,
while returning (besides the recovered state) the query remainder that needs to be evaluated. 
"""
_cache = None


def get_cache():
    """Get global cache object"""
    global _cache
    if _cache is None:
        _cache = NoCache()
    return _cache


def set_cache(cache):
    """Set global cache object"""
    global _cache
    _cache = cache


def cached_part(query, cache=None):
    """Get cached part of the query.
    Use either supplied cache object or global cache object (default).
    In the process, query is into two parts: the beginning of the query
    and the remainder. Function tries to find longest possible beginning of the query
    which is cached, then returns the cached state and the remainder of the query.
    (query == state.query + "/" + remainder)
    """
    if cache is None:
        cache = get_cache()

    if isinstance(
        cache, NoCache
    ):  # Just an optimization - to avoid looping over all query splits
        return State(), encode(decode(query))

    for key, remainder in all_splits(query):
        if key == "":
            return State(), remainder
        if cache.contains(key):
            state = cache.get(key)
            if state is None:
                continue
            return state, remainder

    # Should never get here, but this is a sensible default:
    return State(), encode(decode(query))


class CacheMixin:
    """Adds various cache combinator helpers"""

    def __add__(self, cache):
        return CacheCombine(self, cache)

    def if_contains(self, *attributes):
        "Cache if state contains all attributes"
        return CacheIfHasAttributes(self, *attributes)

    def if_not_contains(self, *attributes):
        "Cache if state contains none of the attributes"
        return CacheIfHasNotAttributes(self, *attributes)

    def if_attribute_equal(self, attribute, value):
        "Cache if state attribute is equal to value"
        return CacheAttributeCondition(self, attribute, value, True)

    def if_attribute_not_equal(self, attribute, value):
        "Cache if state attribute is not equal to value"
        return CacheAttributeCondition(self, attribute, value, False)


class CacheCombine(CacheMixin):
    def __init__(self, cache1, cache2):
        self.cache1 = cache1
        self.cache2 = cache2

    def clean(self):
        self.cache1.clean()
        self.cache2.clean()

    def get(self, key):
        value = self.cache1.get(key)
        if value is not None:
            return value
        else:
            return self.cache2.get(key)

    def get_metadata(self, key):
        value = self.cache1.get_metadata(key)
        if value is not None:
            return value
        else:
            return self.cache2.get_metadata(key)

    def store(self, state):
        self.remove(state.query)
        if self.cache1.store(state):
            return True
        else:
            return self.cache2.store(state)

    def store_metadata(self, metadata):
        if self.cache1.store_metadata(metadata):
            return True
        else:
            return self.cache2.store_metadata(metadata)

    def remove(self, key):
        return self.cache1.remove(key) and self.cache2.remove(key)

    def contains(self, key):
        if self.cache1.contains(key):
            return True
        else:
            return self.cache2.contains(key)

    def keys(self):
        for key in self.cache1.keys():
            yield key
        for key in self.cache2.keys():
            yield key

    def __str__(self):
        return f"{str(self.cache1)} + {str(self.cache2)}"

    def __repr__(self):
        return f"{repr(self.cache1)} + {repr(self.cache2)}"


class CacheIfHasAttributes(CacheMixin):
    def __init__(self, cache, *attributes):
        self.cache = cache
        self.attributes = attributes

    def clean(self):
        self.cache.clean()

    def get(self, key):
        return self.cache.get(key)

    def get_metadata(self, key):
        return self.cache.get_metadata(key)

    def store(self, state):
        self.remove(state.query)
        if all(
            state.metadata.get("attributes", {}).get(a, False) for a in self.attributes
        ):
            return self.cache.store(state)
        else:
            return False

    def store_metadata(self, metadata):
        if all(metadata.get("attributes", {}).get(a, False) for a in self.attributes):
            return self.cache.store_metadata(metadata)
        else:
            return False

    def remove(self, key):
        return self.cache.remove(key)

    def contains(self, key):
        return self.cache.contains(key)

    def keys(self):
        return self.cache.keys()

    def __str__(self):
        return f"({str(self.cache)} containing {', '.join(self.attributes)})"

    def __repr__(self):
        return f"CacheIfHasAttributes({repr(self.cache)}, {', '.join(map(repr,self.attributes))})"


class CacheIfHasNotAttributes(CacheMixin):
    def __init__(self, cache, *attributes):
        self.cache = cache
        self.attributes = attributes

    def clean(self):
        self.cache.clean()

    def get(self, key):
        return self.cache.get(key)

    def get_metadata(self, key):
        return self.cache.get_metadata(key)

    def store(self, state):
        self.remove(state.query)
        if any(
            state.metadata.get("attributes", {}).get(a, False) for a in self.attributes
        ):
            return False
        else:
            return self.cache.store(state)

    def store_metadata(self, metadata):
        if any(metadata.get("attributes", {}).get(a, False) for a in self.attributes):
            return False
        else:
            return self.cache.store_metadata(metadata)

    def remove(self, key):
        return self.cache.remove(key)

    def contains(self, key):
        return self.cache.contains(key)

    def keys(self):
        return self.cache.keys()

    def __str__(self):
        return f"({str(self.cache)} not containing {', '.join(self.attributes)})"

    def __repr__(self):
        return f"CacheIfHasNotAttributes({repr(self.cache)}, {', '.join(map(repr,self.attributes))})"


class CacheAttributeCondition(CacheMixin):
    def __init__(self, cache, attribute, value, equals=True):
        self.cache = cache
        self.attribute = attribute
        self.value = value
        self.equals = equals

    def clean(self):
        self.cache.clean()

    def get(self, key):
        return self.cache.get(key)

    def get_metadata(self, key):
        return self.cache.get_metadata(key)

    def store(self, state):
        self.remove(state.query)
        state_attribute_value = state.metadata.get("attributes", {}).get(self.attribute)
        if self.equals:
            if state_attribute_value == self.value:
                return self.cache.store(state)
        else:
            if state_attribute_value != self.value:
                return self.cache.store(state)
        return False

    def store_metadata(self, metadata):
        state_attribute_value = metadata.get("attributes", {}).get(self.attribute)
        if self.equals:
            if state_attribute_value == self.value:
                return self.cache.store_metadata(metadata)
        else:
            if state_attribute_value != self.value:
                return self.cache.store_metadata(metadata)
        return False

    def remove(self, key):
        return self.cache.remove(key)

    def contains(self, key):
        return self.cache.contains(key)

    def keys(self):
        return self.cache.keys()

    def __str__(self):
        if self.equals:
            return f"({str(self.cache)} when {self.attribute}=={self.value})"
        else:
            return f"({str(self.cache)} when {self.attribute}!={self.value})"

    def __repr__(self):
        return f"CacheAttributeCondition({repr(self.cache)}, {repr(self.attribute)}, {repr(self.value)}, {self.equals})"


class CacheProxy:
    def __init__(self, cache, verbose=False):
        self.cache = cache
        self.verbose = verbose

    def clean(self):
        if self.verbose:
            print("(CacheProxy) clean()")
        self.cache.clean()

    def get(self, key):
        if self.verbose:
            print(f"(CacheProxy) get({key})")
        return self.cache.get(key)

    def get_metadata(self, key):
        if self.verbose:
            print(f"(CacheProxy) get_metadata({key})")
        return self.cache.get_metadata(key)

    def store(self, state):
        if self.verbose:
            print(f"(CacheProxy) store()")
        return self.cache.store(state)

    def store_metadata(self, metadata):
        if self.verbose:
            print(f"(CacheProxy) store_metadata()")
        return self.cache.store_metadata(metadata)

    def remove(self, key):
        if self.verbose:
            print(f"(CacheProxy) remove({key})")
        return self.cache.remove(key)

    def contains(self, key):
        if self.verbose:
            print(f"(CacheProxy) contains({key})")
        return self.cache.contains(key)

    def keys(self):
        if self.verbose:
            print(f"(CacheProxy) keys()")
        return list(self.cache.keys())

    def __str__(self):
        return f"CacheProxy of {str(self.cache)}"

    def __repr__(self):
        return f"CacheProxy({repr(self.cache)})"


class NoCache(CacheMixin):
    """Trivial cache object which does not cache any state"""

    def clean(self):
        pass

    def get(self, key):
        return None

    def get_metadata(self, key):
        return None

    def store(self, state):
        return False

    def store_metadata(self, state):
        return False

    def remove(self, key):
        return False

    def contains(self, key):
        return False

    def keys(self):
        return []

    def __str__(self):
        return "No cache"

    def __repr__(self):
        return "NoCache()"


class MemoryCache(CacheMixin):
    """Simple cache which stores all the states in a dictionary.
    Note that continuous heavy use of the system with MemoryCache
    may lead to filling the memory, therefore this is not ideal
    for long running services.
    """

    def __init__(self):
        self.storage = {}

    def clean(self):
        self.storage = {}

    def get(self, key):
        state = self.storage.get(key)

        if state is None:
            return None
        else:
            if state.metadata.get("status") != "ready":
                return None
            return state.clone()

    def get_metadata(self, key):
        state = self.storage.get(key)
        if state is None:
            return None
        else:
            return dict(**state.metadata)

    def store(self, state):
        if state.is_error:
            return None
        state.metadata["status"] = "ready"
        self.storage[state.query] = state.clone()
        return True

    def store_metadata(self, metadata):
        key = metadata["query"]
        self.storage[key] = self.storage.get(key, State())
        self.storage[key].metadata = metadata

        return True

    def remove(self, key):
        if key in self.storage:
            del self.storage[key]
        return True

    def contains(self, key):
        return key in self.storage

    def keys(self):
        return self.storage.keys()

    def __str__(self):
        return "Memory cache"

    def __repr__(self):
        return "MemoryCache()"


class FileCache(CacheMixin):
    """Simple file cache which stores all the states in files
    in a specified directory of a local filesystem.
    Two files are created: one for the state metadata and the other one with
    serialized version of the state data.

    Note that no mechanism for maintaining freshness or constraining file size
    is provided. This may lead to filling the space on the filesystem,
    therefore this is not ideal for long running public web-services.
    """

    def __init__(self, path):
        self.path = path
        try:
            makedirs(path)
        except FileExistsError:
            pass

    def clean(self):
        import glob

        print(f"Clean {self}")
        for f in glob.glob(os.path.join(self.path, "*")):
            logging.debug(f"Removing cache file {f}")
            os.remove(f)

    def to_path(self, key, prefix="state_", extension="json"):
        "Construct file path from a key and optionally prefix and file extension."
        m = hashlib.md5()
        m.update(key.encode("utf-8"))
        digest = m.hexdigest()
        return os.path.join(self.path, f"{prefix}{digest}.{extension}")

    def encode(self, b):
        return b

    def decode(self, s):
        return s

    def encode_metadata(self, b):
        return self.encode(b.encode("utf-8"))

    def decode_metadata(self, s):
        s = self.decode(s)
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            return s.decode("utf-8")
        else:
            raise Exception(f"Unsupported type: {type(s)}")

    def get(self, key):
        metadata = self.get_metadata(key)
        if metadata is None:
            print(f"(FileCache) Metadata missing: {key}")
            return None
        if metadata.get("status") != "ready":
            print(f"(FileCache) Not ready {key}; ", metadata.get("status"))
            return None
        state = State()
        state.metadata = metadata

        t = state_types_registry().get(metadata["type_identifier"])
        path = self.to_path(key, prefix="data_", extension=t.default_extension())
        if os.path.exists(path):
            try:
                state.data = t.from_bytes(self.decode(open(path, "rb").read()))
                return state
            except:
                traceback.print_exc()
                logging.exception(f"Cache failed to recover {key}")
                return None

    def _load_metadata(self, state_path):
        if os.path.exists(state_path):
            try:
                return json.loads(self.decode_metadata(open(state_path, "rb").read()))
            except:
                traceback.print_exc()
                return None
        else:
            return None

    def get_metadata(self, key):
        return self._load_metadata(self.to_path(key))

    def remove(self, key):
        metadata = self.get_metadata(key)
        if metadata is None:
            return True
        if "type_identifier" in metadata:
            t = state_types_registry().get(metadata["type_identifier"])
            path = self.to_path(key, prefix="data_", extension=t.default_extension())
            if os.path.exists(path):
                os.remove(path)

        state_path = self.to_path(key)
        if os.path.exists(state_path):
            os.remove(state_path)

        return True

    def contains(self, key):
        state_path = self.to_path(key)
        if os.path.exists(state_path):
            state = State()
            metadata = self._load_metadata(state_path)
            if metadata is None:
                return False
            else:
                return metadata.get("query") == key
        else:
            return False
        return True

    #        t = state_types_registry().get(state.type_identifier)
    #        path = self.to_path(key, prefix="data_", extension=t.default_extension())
    #        if os.path.exists(path):
    #            return True
    #        else:
    #            return False

    def keys(self):
        import glob

        for f in glob.glob(os.path.join(self.path, "state_*.json")):
            metadata = self._load_metadata(f)
            if metadata is not None:
                yield metadata["query"]

    def store(self, state):
        if state.is_error:
            return None
        state.metadata["status"] = "ready"

        if not self.store_metadata(state.metadata):
            return False

        t = state_types_registry().get(state.type_identifier)
        path = self.to_path(
            state.query, prefix="data_", extension=t.default_extension()
        )
        with open(path, "wb") as f:
            try:
                b, mime = t.as_bytes(state.data)
                f.write(self.encode(b))
            except NotImplementedError:
                return False
        return True

    def store_metadata(self, metadata):
        try:
            with open(self.to_path(metadata["query"]), "wb") as f:
                f.write(self.encode_metadata(json.dumps(metadata)))
        except:
            logging.exception(f"Cache writing error: {metadata['query']}")
            return False
        return True

    def __str__(self):
        return f"File cache at {self.path}"

    def __repr__(self):
        return f"FileCache('{self.path}')"


class StoreCache(CacheMixin):
    """Simple cache similar to FileCache, but using a store module instead of a local filesystem."""

    def __init__(self, store, path, flat=False):
        self.storage = store
        self.path = path
        if not self.storage.is_dir(path):
            self.storage.makedir(path)
        self.flat = flat

    def clean(self):
        import glob

        print(f"Clean {self}")
        path = "" if self.path in (None, "") else self.path + "/"
        for key in self.storage.keys():
            if not self.storage.is_dir(key) and key.startswith(path):
                logging.debug(f"Removing cache file {key}")
                self.storage.remove(key)
        for _, key in sorted((-len(k.split("/")), k) for k in self.storage.keys()):
            if self.storage.is_dir(key) and key.startswith(path):
                try:
                    logging.debug(f"Removing cache dir {key}")
                    self.storage.removedir(key)
                except:
                    logging.debug(f"Failed to remove cache dir {key}")

    def to_path(self, key, prefix="0state_"):
        "Construct file path from a key and optionally prefix and file extension."
        if self.flat:
            m = hashlib.md5()
            m.update(key.encode("utf-8"))
            digest = m.hexdigest()
            path = f"{self.path}/{prefix}{digest}.data"
        else:
            path = f"{self.path}/{key}/{prefix}.data"
        if path.startswith("/"):
            path = path[1:]
        return path

    def encode(self, b):
        return b

    def decode(self, s):
        return s

    def encode_metadata(self, b):
        return self.encode(b.encode("utf-8"))

    def decode_metadata(self, s):
        s = self.decode(s)
        if isinstance(s, str):
            return s
        elif isinstance(s, bytes):
            return s.decode("utf-8")
        else:
            raise Exception(f"Unsupported type: {type(s)}")

    def get(self, key):
        print(f"GET {key}")
        metadata = self.get_metadata(key)
        print(f"  METADATA {metadata}")
        if metadata is None:
            print(f"(StoreCache) Metadata missing: {key}")
            return None
        if metadata.get("status") != "ready":
            print(f"(StoreCache) Not ready {key}; ", metadata.get("status"))
            return None
        state = State()
        state.metadata = metadata

        t = state_types_registry().get(metadata["type_identifier"])
        path = self.to_path(key)
        if self.storage.contains(path):
            try:
                state.data = t.from_bytes(self.decode(self.storage.get_bytes(path)))
                return state
            except:
                traceback.print_exc()
                logging.exception(f"Cache failed to recover {key}")
                return None

    def _load_metadata(self, state_path):
        if self.storage.contains(state_path) and not self.storage.is_dir(state_path):
            return self.storage.get_metadata(state_path)

    def get_metadata(self, key):
        return self._load_metadata(self.to_path(key))

    def remove(self, key):
        try:
            self.storage.remove(self.to_path(key))
            return True
        except:
            traceback.print_exc()
            return False

    def contains(self, key):
        state_path = self.to_path(key)
        return self.storage.contains(state_path)

    def keys(self):
        path = self.path + "/"
        for key in self.storage.keys():
            if self.path in ("", None) or key.startswith(path):
                if not self.storage.is_dir(key):
                    metadata = self.storage.get_metadata(key)
                    if "query" in metadata:
                        yield metadata["query"]

    def store(self, state):
        if state.is_error:
            return None
        state.metadata["status"] = "ready"

        t = state_types_registry().get(state.type_identifier)
        path = self.to_path(state.query)
        if self.storage.is_supported(path):
            try:
                b, mime = t.as_bytes(state.data)
                metadata = dict(**state.metadata)
                metadata["mimetype"] = mime
                self.storage.store(path, b, metadata)
                return True
            except:
                return False
        return False

    def store_metadata(self, metadata):
        try:
            key = self.to_path(metadata["query"])
            if self.storage.is_supported(key):
                self.storage.store_metadata(key, metadata)
                return True
        except:
            logging.exception(f"Cache metadata storing error: {metadata['query']}")
            return False
        return False

    def __str__(self):
        return f"Store cache at {self.path} based on {self.storage}"

    def __repr__(self):
        return f"FileCache({repr(self.storage)}, {repr(self.path)}, flat={self.flat})"


class XORFileCache(FileCache):
    def __init__(self, path, code):
        super().__init__(path)
        self.code = np.frombuffer(code, dtype=np.uint8)

    def code_of_length(self, n):
        if n <= len(self.code):
            return self.code[:n]
        else:
            return np.tile(self.code, int(n / len(self.code)) + 1)[:n]

    def encode(self, b):
        ba = np.frombuffer(b, dtype=np.uint8)
        return (ba ^ self.code_of_length(len(b))).tobytes()

    def decode(self, s):
        return self.encode(s)


class FernetFileCache(FileCache):
    def __init__(self, path, fernet_key):
        from cryptography.fernet import Fernet

        super().__init__(path)
        self.fernet = Fernet(fernet_key)

    def encode(self, b):
        return self.fernet.encrypt(b)

    def decode(self, s):
        return self.fernet.decrypt(s)


class SQLCache(CacheMixin):
    """Store cache in a SQL database.
    Tested with sqlite3.
    For databases without BLOB support (e.g. Hive) use SQLStringCache.
    """

    def __init__(
        self,
        connection=None,
        table="liquer_cache",
        metadata_type="TEXT",
        state_data_type="BLOB",
        delete_before_insert=False,
        store_metadata_enabled=True,
    ):
        self.connection = connection
        self.table = table
        self.metadata_type = metadata_type
        self.state_data_type = state_data_type
        self.delete_before_insert = delete_before_insert
        self._available_keys = None
        self.store_metadata_enabled = store_metadata_enabled
        self.init()

    def init(self):
        try:
            query = f"""CREATE TABLE {self.table} (
                query         VARCHAR(2000),
                metadata      {self.metadata_type},
                state_data    {self.state_data_type}
            )
            """
            print(query)
            logging.debug(f"CACHE TABLE: {query}")
            c = self.connection.cursor()
            c.execute(query)
        except:
            traceback.print_exc()

    @classmethod
    def from_sqlite(cls, path=":memory:", table="liquer_cache"):
        import sqlite3

        connection = sqlite3.connect(path)
        return cls(connection=connection, table=table, delete_before_insert=True)

    @property
    def available_keys(self):
        if self._available_keys is None:
            c = self.connection.cursor()
            c.execute(
                f"""
            SELECT
              query
            FROM {self.table}
            """
            )

            try:
                self._available_keys = [x[0] for x in c.fetchall()]
            except:
                traceback.print_exc()
                print("SQL available_keys failed")
                return None
        return self._available_keys

    def clean(self):
        c = self.connection.cursor()
        c.execute(f"""DROP TABLE {self.table}""")
        self._available_keys = []
        self.init()
        self.connection.commit()

    def encode(self, b):
        return b

    def decode(self, s):
        return s

    def get(self, key):
        c = self.connection.cursor()
        c.execute(
            f"""
        SELECT
          metadata,
          state_data
        FROM {self.table}
        WHERE query=?
        """,
            [key],
        )

        try:
            metadata, data = c.fetchone()
            metadata = json.loads(metadata)
            if metadata.get("status") != "ready":
                return None
        except:
            return None
        try:
            state = State()
            state = state.from_dict(metadata)

            t = state_types_registry().get(state.type_identifier)
            state.data = t.from_bytes(self.decode(data))
            return state
        except:
            logging.exception(f"Cache failed to recover {key}")
            return None

    def get_metadata(self, key):
        c = self.connection.cursor()
        c.execute(
            f"""
        SELECT
          metadata
        FROM {self.table}
        WHERE query=?
        """,
            [key],
        )

        try:
            (metadata,) = c.fetchone()
        except:
            return None
        try:
            return json.loads(metadata)
        except:
            logging.exception(f"Cache failed to recover metadata {key}")
            return None

    def contains(self, key):
        return key in self.available_keys

    def keys(self):
        return self.available_keys

    def store(self, state):
        if state.is_error:
            return None
        state.metadata["status"] = "ready"

        key = state.query
        metadata = json.dumps(state.as_dict())

        t = state_types_registry().get(state.type_identifier)
        try:
            b, mime = t.as_bytes(state.data)
        except NotImplementedError:
            return False
        self._available_keys = None
        if self.delete_before_insert:
            self.connection.execute(f"DELETE FROM {self.table} WHERE query=?", [key])
        self.connection.execute(
            f"INSERT INTO {self.table} (query, metadata, state_data) VALUES (?, ?, ?)",
            [key, metadata, self.encode(b)],
        )
        self.connection.commit()
        return True

    def store_metadata(self, metadata):
        if self.store_metadata_enabled:
            key = metadata["query"]
            metadata = json.dumps(metadata)

            self._available_keys = None
            if self.delete_before_insert:
                self.connection.execute(
                    f"DELETE FROM {self.table} WHERE query=?", [key]
                )
            self.connection.execute(
                f"INSERT INTO {self.table} (query, metadata, state_data) VALUES (?, ?, ?)",
                [key, metadata, None],
            )
            self.connection.commit()
            return True
        else:
            return False

    def remove(self, key):
        self.connection.execute(f"DELETE FROM {self.table} WHERE query=?", [key])
        self.connection.commit()
        return True

    def __str__(self):
        return f"SQL cache {self.table}"

    def __repr__(self):
        return f"SQLCache(table='{self.table}')"


class SQLStringCache(SQLCache):
    """Store cache in a SQL database.
    Data are encoded into string before storing in the database.
    This is suitable for databases without BLOB support (e.g. Hive).
    By default base64 encoding is used.
    To change the encoding override encode and decode methods.
    """

    def __init__(
        self,
        connection=None,
        table="liquer_cache",
        text_type="STRING",
        delete_before_insert=False,
    ):
        super().__init__(
            connection=connection,
            table=table,
            metadata_type=text_type,
            state_data_type=text_type,
            delete_before_insert=delete_before_insert,
        )

    @classmethod
    def from_sqlite(cls, path=":memory:", table="liquer_cache"):
        import sqlite3

        connection = sqlite3.connect(path)
        return cls(connection=connection, table=table, text_type="TEXT")

    def encode(self, b):
        return base64.b64encode(b)

    def decode(self, s):
        return base64.b64decode(s)

    def __str__(self):
        return f"SQL string cache {self.table}"

    def __repr__(self):
        return f"SQLStringCache(table='{self.table}')"



from liquer import *
import json
from liquer.commands import command_registry
from liquer.state_types import encode_state_data, state_types_registry
import traceback
import requests
from liquer.query import evaluate
from liquer.state import get_vars
from liquer.cache import get_cache
from liquer.store import get_store, KeyNotFoundStoreException
import io
import traceback


def liquer_static_path():
    import liquer.server
    import os.path

    return os.path.join(os.path.dirname(liquer.server.__file__), "static")


class LiquerIndexHandler:
    def get(self):
        self.redirect("/static/index.html")


class LiquerJsHandler:
    def get(self):
        self.redirect("/static/liquer.js")


# /api/commands.json
class CommandsHandler:
    def get(self):
        """Returns a list of commands in json format"""
        self.write(json.dumps(command_registry().as_dict()))


def response(state):
    """Create flask response from a State"""
    filename = state.metadata.get("filename")
    b, mimetype, type_identifier = encode_state_data(state.get(), extension=state.extension)
    if filename is None:
        filename = state_types_registry().get(type_identifier).default_filename()
    return b, mimetype, filename


#'/submit/<query>
class SubmitHandler:
    """Submit query.
    Starts query in the background.
    """
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)

    def get(self, query):
        from liquer.pool import evaluate_in_background
        evaluate_in_background(query)
        self.write(
            json.dumps(
                dict(status="OK", message="Submitted", query=query)
            )
        )

    def post(self, param):
        self.write(
            json.dumps(command_registry().register_remote_serialized(self.request.body))
        )

# /q/<path:query>
class QueryHandler:
    def get(self, query):
        """Main service for evaluating queries"""        
        b, mimetype, filename = response(evaluate(query))

        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)

        self.write(b)


#/api/cache/get/<path:query>
class CacheGetDataHandler:
    def get(self, query):
        """Main service for evaluating queries"""
        state = get_cache().get(query)
        if state is None:
            self.set_status(404)
            self.finish(f"404 - {query} not found in cache")

        b, mimetype, filename = response(state)
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)

        self.write(b)

#/api/cache/meta/<path:query>
class CacheMetadataHandler:
    def get(self, query):
        metadata = get_cache().get_metadata(query)
        if metadata == False:
            metadata = dict(query=query, status="not available", cached=False)
            self.write(metadata)
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(metadata))
    def post(self, param):
        try:
            metadata = json.loads(self.request.body)
            query = metadata.get('query')
            result_code = get_cache().store_metadata(metadata)
            result = dict(
                query=query, result=result_code, status="OK", message="OK", traceback=""
            )
        except Exception as e:
            result = dict(
                query=query,
                result=False,
                status="ERROR",
                message=str(e),
                traceback=traceback.format_exc(),
            )
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(result))


#/api/cache/remove/<path:query>
class CacheRemoveHandler:
    def get(self, query):
        r = get_cache().remove(query)
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(dict(query=query, removed=r)))


#/api/cache/contains/<path:query>
class CacheContainsHandler:
    def get(self, query):
        contains = get_cache().contains(query)
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(dict(query=query, cached=contains)))


#/api/cache/keys.json
class CacheKeysHandler:
    def get(self, query):
        keys = dict(keys=list(get_cache().keys()))
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(keys))


#/api/cache/clean
class CacheCleanHandler:
    def get(self):
        get_cache().clean()
        n = len(list(get_cache().keys()))
        keys = dict(status="OK", message=f"Cache cleaned, {n} keys left")
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(keys))


#/api/commands.json
class CommandsHandler:
    def get(self):
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(command_registry().as_dict()))


#/api/debug-json/<path:query>
class GetMetadataHandler:
    """Debug query - returns metadata from a state after a query is evaluated"""
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        state = evaluate(query)
        state_json = state.as_dict()
        self.write(json.dumps(state_json))


#/api/stored_metadata/<path:query>
class GetStoredMetadataHandler:
    """Get metadata stored in a store or cache"""
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        import liquer.tools
        metadata = liquer.tools.get_stored_metadata(query)
        self.write(json.dumps(metadata))


# /api/build
class BuildHandler:
    """Build a query from a posted decoded query (list of lists of strings).
    Result is a dictionary with encoded query and link.
    """

    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)

    def post(self):
        from liquer.parser import encode

        query = encode(json.loads(self.request.body)["ql"])
        link = (
            get_vars().get("server", "http://localhost")
            + get_vars().get("api_path", "/q/")
            + query
        )
        self.write(json.dumps(dict(query=query, link=link, message="OK", status="OK")))


#'/api/register_command/
class RegisterCommandHandler:
    """Remote command registration service.
    This has to be enabled by liquer.commands.enable_remote_registration()

    WARNING: Remote command registration allows to deploy arbitrary python code on LiQuer server,
    therefore it is a HUGE SECURITY RISK and it only should be used if other security measures are taken
    (e.g. on localhost or intranet where only trusted users have access).
    This is on by default on Jupyter server extension.
    """

    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)

    def get(self, param):
        self.write(
            json.dumps(
                command_registry().register_remote_serialized(param.encode("ascii"))
            )
        )

    def post(self, param):
        self.write(
            json.dumps(command_registry().register_remote_serialized(self.request.body))
        )

#/api/store/data/<path:query>
class StoreDataHandler:
    def get(self, query):
        """Get data from store. Equivalent to Store.get_bytes.
        Content type (MIME) is obtained from the metadata.
        """
        store = get_store()
        metadata = store.get_metadata(query)

        try:
            mimetype = metadata.get("mimetype", "application/octet-stream")
            b=store.get_bytes(query)
        except:
            mimetype = "application/json"
            b = json.dumps(
                dict(query=query, message=traceback.format_exc(), status="ERROR")
            )
 
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(b)

    def post(self, query):
        """Set data from store. Equivalent to Store.store.
        Unlike store method, which stores both data and metadata in one call,
        the api/store/data POST only stores the data. The metadata needs to be set in a separate POST of api/store/metadata
        either before or after the api/store/data POST.
        """
        store = get_store()
        try:
            metadata = store.get_metadata(query)
        except KeyNotFoundStoreException:
            metadata={}
        try:        
            data = self.request.body
            store.store(query, data, metadata)
            response = dict(query=query, message="Data stored", status="OK")
        except:
            response = dict(query=query, message=traceback.format_exc(), status="ERROR")

        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(response))

#/api/store/metadata/<path:query>
class StoreMetadataHandler:
    def get(self, query):
        """Get data from store. Equivalent to Store.get_bytes.
        Content type (MIME) is obtained from the metadata.
        """
        store = get_store()
        metadata = store.get_metadata(query)
 
        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(metadata))

    def post(self, param):
        """Set data from store. Equivalent to Store.store.
        Unlike store method, which stores both data and metadata in one call,
        the api/store/data POST only stores the data. The metadata needs to be set in a separate POST of api/store/metadata
        either before or after the api/store/data POST.
        """
        store = get_store()
        try:
            metadata = json.loads(self.request.body)
            query = metadata.get('query')
            store.store_metadata(query, metadata)
            response = dict(query=query, message="Metadata stored", status="OK")
        except:
            response = dict(query=query, message=traceback.format_exc(), status="ERROR")

        mimetype = "application/json"
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(json.dumps(response))

#/web/<path:query>
class WebStoreHandler:
    def get(self, query):
        """Shortcut to the 'web' directory in the store.
        Similar to /store/data/web, except the index.html is automatically added if query is a directory.
        The 'web' directory hosts web applications and visualization tools, e.g. liquer-pcv or liquer-gui. 
        """
        store = get_store()

        try:
            query="web/"+query
            if query.endswith("/"):
                query+="index.html"
            if store.is_dir(query):
                query+="/index.html"
            metadata = store.get_metadata(query)
            mimetype = metadata.get("mimetype", "application/octet-stream")
            b=store.get_bytes(query)
        except:
            mimetype = "application/json"
            b = json.dumps(
                dict(query=query, message=traceback.format_exc(), status="ERROR")
            )
 
        header = "Content-Type"
        body = mimetype
        self.set_header(header, body)
        self.write(b)


#/api/store/remove/<path:query>
class StoreRemoveHandler:
    """Handler to remove data from store"""
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            store.remove(query)
            self.write(json.dumps(dict(query=query, message=f"Removed {query}", status="OK")))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/removedir/<path:query>
class StoreRemovedirHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            store.removedir(query)
            self.write(json.dumps(dict(query=query, message=f"Removed directory {query}", status="OK")))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/contains/<path:query>
class StoreContainsHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            contains = store.contains(query)
            self.write(json.dumps(
                dict(
                    query=query, message=f"Contains {query}", contains=contains, status="OK"
                )
            ))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/is_dir/<path:query>
class StoreIsDirHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            is_dir = store.is_dir(query)
            self.write(json.dumps(
                dict(
                    query=query, message=f"Is directory {query}", is_dir=is_dir, status="OK"
                )
            ))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/keys")
class StoreKeysHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            keys = store.keys()
            self.write(json.dumps(
                dict(query=None, message=f"Keys obtained", keys=keys, status="OK")
            ))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/listdir/<path:query>
class StoreListdirHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            listdir = store.listdir(query)
            self.write(json.dumps(
                dict(query=query, message=f"Keys obtained", listdir=listdir, status="OK")
            ))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))


#/api/store/makedir/<path:query>
class StoreMakedirHandler:
    def prepare(self):
        header = "Content-Type"
        body = "application/json"
        self.set_header(header, body)
    def get(self, query):
        store = get_store()
        try:
            store.makedir(query)
            self.write(json.dumps(
                dict(query=query, message=f"Makedir succeeded", status="OK")
            ))
        except:
            self.write(json.dumps(dict(query=query, message=traceback.format_exc(), status="ERROR")))




from io import StringIO, BytesIO
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

from liquer.commands import command, first_command
from liquer.context import get_context
import pandas as pd
import polars as pl

class PolarsDataframeStateType(StateType):
    def identifier(self):
        return "polars_dataframe"

    def default_extension(self):
        return "parquet"

    def is_type_of(self, data):
        return isinstance(data, pl.DataFrame)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)

        if extension == "csv":
            output = BytesIO()
            data.to_csv(output)
            return output.getvalue(), mimetype
        elif extension == "parquet":
            output = BytesIO()
            data.to_parquet(output)
            return output.getvalue(), mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by polars data-frame type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()
        f = BytesIO()
        f.write(b)
        f.seek(0)
        if extension == "csv":
            return pl.read_csv(f)
        elif extension == "parquet":
            return pl.read_parquet(f)
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by polars data-frame type."
        )

    def copy(self, data):
        return self.from_bytes(self.as_bytes(data)[0])

    def data_characteristics(self, data):
        return dict(description=f"Polars data-frame with {len(data.columns)} and {len(data)} rows.")
        

POLARS_DATAFRAME_STATE_TYPE = PolarsDataframeStateType()
register_state_type(pl.DataFrame, POLARS_DATAFRAME_STATE_TYPE)

@command
def polars_df(data, extension=None, context=None):
    """Convert bytes or a dataframe to a workbook"""
    context=get_context(context)
    if type(data)==bytes:
        context.info(f"Polars data-frame from bytes. Extension:'{extension}'")
        return POLARS_DATAFRAME_STATE_TYPE.from_bytes(data, extension=extension)
    elif isinstance(data,pd.DataFrame):
        context.info("Polars data-frame from Pandas data-frame")
        return pl.DataFrame(data)
    elif isinstance(data,pl.DataFrame):
        context.info("Polars data-frame kept as it is")
        return data
    raise Exception(f"Unsupported polars dataframe type: {type(data)}")



from liquer import *
import pandas as pd
from liquer.store import get_store
from liquer.state_types import (
    data_characteristics,
    state_types_registry,
    StateType,
    register_state_type,
)
from liquer.constants import mimetype_from_extension
import json


@command
def concat_batches(idf, max_batches=None, context=None):
    """Concatenates an iterator of dataframes (batches) into a single dataframe
    A maximum limit of batches to be processed can be specified with max_batches.
    By default all the batches will be concatenated.
    """
    context = get_context(context)
    context.info(f"Concatenate batches")
    if isinstance(idf, pd.DataFrame):
        context.info(f"Single dataframe received")
        idf = [idf]
    batch_number = 0
    if max_batches in ("0", "", None):
        max_batches = 0
    else:
        max_batches = int(max_batches)
    buffer = pd.DataFrame()
    for df in context.progress_iter(idf):
        if not len(df):
            continue
        context.info(
            f"Receiving dataframe with {len(df)} rows and {len(df.columns)} columns"
        )
        if len(buffer) > 0:
            if len(df.columns) != len(buffer.columns):
                context.warning(
                    f"Number of columns in the batches differs - before:{len(buffer.columns)}, now:{len(df.columns)}"
                )
        batch_number += 1
        if max_batches:
            context.info(f"Concatenate batch {batch_number}/{max_batches}")
        else:
            context.info(f"Concatenate batch {batch_number}")
        buffer = buffer.append(df, ignore_index=True)
        if max_batches and batch_number >= max_batches:
            context.info(f"Maximum number of batches reached")
            break
    return buffer


@command(volatile=True, cache=False)
def repackage_batches(idf, batch_size=1024, max_batches=0, context=None):
    """Repackages an iterator of dataframes (batches) into fixed-length batches.
    Results in an iterator of dataframes where each dataframe (with the exception of the last one)
    has a the batch_size number of rows.
    Note: Since the result is an iterator, it is volatile and can't be cached.
    A maximum limit of batches to be processed can be specified with max_batches.
    By default all the batches will be repackaged.
    """
    context = get_context(context)
    context.info(f"Repackage batches")
    if isinstance(idf, pd.DataFrame):
        context.info(f"Single dataframe received")
        idf = [idf]
    buffer = pd.DataFrame()
    if max_batches in ("0", "", None):
        max_batches = 0
    else:
        max_batches = int(max_batches)
    batch_number = 0
    for df in context.progress_iter(idf):
        if not len(df):
            continue
        context.info(
            f"Receiving dataframe with {len(df)} rows and {len(df.columns)} columns"
        )
        if len(buffer) > 0:
            if len(df.columns) != len(buffer.columns):
                context.warning(
                    f"Number of columns in the batches differs - before:{len(buffer.columns)}, now:{len(df.columns)}"
                )
        buffer = buffer.append(df, ignore_index=True)
        while len(buffer) >= batch_size:
            batch_number += 1
            if max_batches:
                context.info(f"Yield batch {batch_number}/{max_batches}")
            else:
                context.info(f"Yield batch {batch_number}")
            result = buffer.iloc[:batch_size, :]
            buffer = buffer.iloc[batch_size:, :]
            yield result
            if max_batches and batch_number >= max_batches:
                context.info(f"Maximum number of batches reached")
                return
    while len(buffer) >= batch_size:
        batch_number += 1
        if max_batches:
            context.info(f"Yield batch {batch_number}/{max_batches} (finishing)")
        else:
            context.info(f"Yield batch {batch_number} (finishing)")
        result = buffer.iloc[:batch_size, :]
        buffer = buffer.iloc[batch_size:, :]
        yield result
        if max_batches and batch_number >= max_batches:
            context.info(f"Maximum number of batches reached")
            return
    if len(buffer):
        batch_number += 1
        if max_batches:
            context.info(f"Yield batch {batch_number}/{max_batches} (last)")
        else:
            context.info(f"Yield batch {batch_number} (last)")
        yield buffer


class StoredDataframeIterator(object):
    def __init__(
        self,
        key,
        item_keys=None,
        extension="parquet",
        number_format="%04d",
        batch_number=0,
        store=None,
    ):
        self.key = key
        self.item_keys = item_keys or []
        self.extension = extension
        self.number_format = number_format
        self.batch_number = batch_number
        self.state_type = state_types_registry().get("dataframe")

        if store is None:
            store = get_store()
        self.store= store

    def to_dict(self, with_batch_number=False):
        d = dict(
            key=self.key,
            item_keys=self.item_keys[:],
            extension=self.extension,
            number_format=self.number_format,
        )
        if with_batch_number:
            d["batch_number"] = self.batch_number
        return d

    @classmethod
    def from_dict(cls, d):
        return cls(
            d["key"],
            item_keys=d.get("item_keys", []),
            extension=d.get("extension"),
            number_format=d.get("number_format"),
            batch_number=d.get("batch_number", 0),
        )

    def copy(self):
        return StoredDataframeIterator(
            self.key,
            item_keys=self.item_keys,
            extension=self.extension,
            number_format=self.number_format,
            batch_number=self.batch_number,
            store=self.store,
        )

    def _key_to_value(self, key):
        if not self.store.contains(key):
            raise Exception(f"Batch {self.batch_number} failure: '{key}' not in store")
        if self.store.is_dir(key):
            raise Exception(
                f"Batch {self.batch_number} failure: '{key}' is a directory"
            )
        metadata = self.store.get_metadata(key)
        if metadata is None:
            print(f"WARNING: Batch {self.batch_number}, key '{key}' missing metadata")
        b = self.store.get_bytes(key)
        if b is None:
            raise Exception(f"Batch {self.batch_number} failure: '{key}' has no data")
        assert type(b) == bytes

        v = key.split(".")
        extension = self.extension if len(v) <= 1 else v[-1]

        df = self.state_type.from_bytes(b, extension=extension)

        return df

    def new_key(self):
        batch_number = len(self.item_keys) + 1
        if self.number_format is None:
            name = str(batch_number)
        else:
            name = self.number_format % batch_number
        if self.extension in (None, ""):
            key = f"{self.key}/{name}"
        else:
            key = f"{self.key}/{name}.{self.extension}"
        if key in self.item_keys:
            raise Exception(f"New key '{key}' already exists")
        return key

    def append(self, df):
        key = self.new_key()
        store = self.store
        b, mimetype = self.state_type.as_bytes(df, self.extension)
        dc = data_characteristics(df)
        assert dc["type_identifier"] == "dataframe"
        metadata = dict(
            type_identifier=dc.get("type_identifier"),
            data_characteristics=data_characteristics(df),
        )
        store.store(key, b, metadata)
        self.item_keys.append(key)

    def rewind(self):
        self.batch_number = 0
        return self

    def __len__(self):
        return len(self.item_keys)

    def __iter__(self):
        return self.copy().rewind()

    def __next__(self):
        if len(self.item_keys) > self.batch_number:
            key = self.item_keys[self.batch_number]
            self.batch_number += 1
            return self._key_to_value(key)
        raise StopIteration

    def __str__(self):
        return f"Dataframe iterator stored in {self.key}"

    def __repr__(self):
        return f"StoredDataframeIterator('{self.key}')"

class StoredDataframeIteratorStateType(StateType):
    def identifier(self):
        return "dataframe_iterator"

    def default_extension(self):
        return "idf"

    def is_type_of(self, data):
        return isinstance(data, StoredDataframeIterator)

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        if extension in ["idf", "json"]:
            mimetype = mimetype_from_extension("json")
            d = data.to_dict()
            return json.dumps(d).encode("utf-8"), mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by stored dataframe iterator type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()

        if extension in ["idf", "json"]:
            return StoredDataframeIterator.from_dict(json.loads(b.decode("utf-8")))
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by dataframe type."
        )

    def copy(self, data):
        return data.copy()

    def data_characteristics(self, data):
        return dict(
            description=f"Dataframe iterator with {len(data.item_keys)} batches."
        )


STORED_DATAFRAME_ITERATOR_STATE_TYPE = StoredDataframeIteratorStateType()
register_state_type(StoredDataframeIterator, STORED_DATAFRAME_ITERATOR_STATE_TYPE)


def _store_batches(idf, key, max_batches=None, context=None):
    """Store iterator of dataframes (batches) in a store.
    The key specifies a directory in the store where the items will be stored.
    Helper function yielding StoredDataframeIterator object and dataframes.
    """
    context = get_context(context)
    context.info(f"Store iterator")
    batch_number = 0
    if max_batches in ("0", "", None):
        max_batches = 0
    else:
        max_batches = int(max_batches)
    store = context.store()
    if store.contains(key):
        if store.is_dir(key):
            context.info(f"Cleaning {key}")
            for x in store.listdir_keys(key):
                context.info(f"Remove {x}")
                store.remove(x)
        else:
            raise Exception(
                f"Can't store the iterator in '{key}'. The key exists and it is not a directory."
            )
    sdfi_key = store.join_key(key, "dataframe_iterator.json")

    sdfi = StoredDataframeIterator(key)
    for df in context.progress_iter(idf):
        if not len(df):
            continue
        batch_number += 1
        if max_batches:
            context.info(f"Storing batch {batch_number}/{max_batches}")
        else:
            context.info(f"Storing batch {batch_number}")
        sdfi.append(df)
        context.store_data(sdfi_key, sdfi)
        yield sdfi, df
#        sdfi_bytes, mimetype = STORED_DATAFRAME_ITERATOR_STATE_TYPE.as_bytes(sdfi)
#        dc = data_characteristics(sdfi)
#        sdfi_metadata = context.metadata()
#        sdfi_metadata.update(
#            dict(type_identifier=dc["type_identifier"], data_characteristics=dc)
#        )
#        store.store(sdfi_key, sdfi_bytes, sdfi_metadata)
        if max_batches and batch_number > max_batches:
            context.info(f"Maximum number of batches reached")
            break
@command
def store_batches(idf, key, max_batches=None, context=None):
    """Store iterator of dataframes (batches) in a store.
    The key specifies a directory in the store where the items will be stored.
    Results in a StoredDataframeIterator object after all batches have been processed.
    This object can be serialized or stored in the object.
    After each iteration step, the StoredDataframeIterator is stored as well in key/dataframe_iterator.json as a side-effect.
    Thus for long-running iterations, the partial data is stored even if the evaluation does not finish.
    """
    context = get_context(context)
    sdfi = StoredDataframeIterator(key)
    for sdfi, df in _store_batches(idf, key, max_batches=max_batches, context=context):
        pass
    return sdfi

@command(cache=False, volatile=True)
def store_batches_pass_through(idf, key, max_batches=None, context=None):
    """Store iterator of dataframes (batches) in a store.
    The key specifies a directory in the store where the items will be stored.
    Unlike store_batches, this immediately yields the dataframes,
    thus the result is a volatile iterator which can not be stored in cache.
    The StoredDataframeIterator is, however, stored in key/dataframe_iterator.json as a side-effect after each iteration step.
    """
    context = get_context(context)
    for sdfi, df in _store_batches(idf, key, max_batches=max_batches, context=context):
        yield df



from liquer import *
from tempfile import TemporaryDirectory
from pathlib import Path

@command
def sweetviz_analyze(df, context=None):
    import sweetviz as sv
    context = get_context(context)
    context.info(f"Sweetviz analyze")
    report = sv.analyze(df)
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)/"sweetviz_analyze_report.html"
        context.info(f"Using temporary file {path}")
        report.show_html(filepath=str(path), open_browser=False)
        with path.open() as f:
            return f.read()



from liquer.commands import command, first_command
import pandas as pd
import numpy as np


def label_columns(df):
    return [c for c in df.columns if len(str(c)) and str(c)[0].isupper()]


def feature_columns(df):
    return [c for c in df.columns if len(str(c)) and str(c)[0].islower()]


def default_ycolumn(df):
    ycolumns = label_columns(df)
    if len(ycolumns):
        return ycolumns[-1]
    else:
        return df.columns[-1]


def to_numpy(df):
    try:
        return df.to_numpy()
    except:
        return df.values


@command(ns="sklearn")
def linear_regression(df, ycolumn=None, fit_intercept=True, normalize=False):
    from sklearn.linear_model import LinearRegression

    if ycolumn is None:
        ycolumn = default_ycolumn(df)
    assert ycolumn in df.columns

    xcolumns = feature_columns(df)
    X = to_numpy(df[xcolumns])
    Y = to_numpy(df[ycolumn]).flatten()
    fit = LinearRegression(fit_intercept=fit_intercept, normalize=normalize).fit(X, Y)
    coefficients = list(fit.coef_)
    model_df = pd.DataFrame([coefficients], columns=xcolumns)
    if fit_intercept:
        model_df.loc[:, "intercept"] = fit.intercept_
    return model_df


@command(ns="sklearn")
def ridge(df, alpha=0.1, ycolumn=None, fit_intercept=True, normalize=False):
    from sklearn.linear_model import Ridge

    if ycolumn is None:
        ycolumn = default_ycolumn(df)
    assert ycolumn in df.columns

    xcolumns = feature_columns(df)
    X = to_numpy(df[xcolumns])
    Y = to_numpy(df[ycolumn]).flatten()
    fit = Ridge(alpha=alpha, fit_intercept=fit_intercept, normalize=normalize).fit(X, Y)
    coefficients = list(fit.coef_)
    model_df = pd.DataFrame([coefficients], columns=xcolumns)
    if fit_intercept:
        model_df.loc[:, "intercept"] = fit.intercept_
    return model_df



from io import BytesIO

from liquer.commands import command
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

import PIL

class PILImageStateType(StateType):
    def identifier(self):
        return "image"

    def default_extension(self):
        return "png"

    def is_type_of(self, data):
        return isinstance(data, PIL.Image.Image)

    def format_from_extension(self, extension):
        if extension is None:
            extension = self.default_extension()
        if extension.lower() in ("bmp", "eps", "jpeg", "pcx", "png", "ppm", "tga", "tiff", "xbm", "ico", "gif"):
            return extension.upper(), True, True
        if extension.lower() == "jpg":
            return "JPEG", True, True
        if extension.lower() in ("pcd", "psd", "wmf"):
            return extension.upper(), True, False
        if extension.lower() == "pdf":
            return extension.upper(), False, True

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)
        format_name, can_read, can_write = self.format_from_extension(extension)
        if can_write: 
            output = BytesIO()
            data.save(output, format=format_name)
            return output.getvalue(), mimetype
        else:
            if can_read:
                raise Exception(
                    f"Serialization: PIL Image only supports reading, but not writing for file extension {extension}."
                )
            else:
                raise Exception(
                    f"Serialization: file extension {extension} is not supported by PIL Image."
                )

    def from_bytes(self, b: bytes, extension=None):
        format_name, can_read, can_write = self.format_from_extension(extension)
        if extension is None:
            formats=None
        else:
            formats=[format_name]

        if can_read:
            f = BytesIO()
            f.write(b)
            f.seek(0)
            return PIL.Image.open(f, formats=formats)
        else:
            if can_write:
                raise Exception(
                    f"Deserialization: PIL Image only supports writing, but not reading for file extension {extension}."
                )
            else:
                raise Exception(
                    f"Deserialization: file extension {extension} is not supported by PIL Image type."
                )

    def copy(self, data):
        return data.copy()

    def data_characteristics(self, data):
        width, height = data.size
        return dict(description=f"Image {width}x{height}")


PIL_IMAGE_STATE_TYPE = PILImageStateType()
register_state_type(PIL.Image.Image, PIL_IMAGE_STATE_TYPE)


@command(ns="pil")
def resize(image, width, height, resample=None):
    """Resize image"""
    resample = dict(
        nearest=PIL.Image.NEAREST,
        box=PIL.Image.BOX,
        bilinear=PIL.Image.BILINEAR,
        hamming=PIL.Image.HAMMING,
        bicubic=PIL.Image.BICUBIC,
        lanczos=PIL.Image.LANCZOS,
    ).get(str(resample).lower())     
    
    return image.copy().resize((int(width), int(height)), resample=resample)



import io
from urllib.request import urlopen
import base64
from liquer.commands import command, first_command
from liquer.parser import encode, decode
from liquer.state_types import encode_state_data
from liquer.cache import get_cache
from liquer.constants import Status, type_identifier_from_extension
from liquer.context import get_context

@command
def let(state, name, value):
    """Set the state variable
    Sets a state variable with 'name' to 'value'.
    """
    state.vars[name] = value
    return state


@command
def flag(state, name, value: bool = True):
    """Set the state variable as a flag (boolean value)"""
    state.vars[name] = bool(value)
    return state


@command
def state_variable(state, name):
    """Get the state variable value
    Useful mainly for debugging and testing
    """
    return state.with_data(state.vars.get(name))


@command
def link(state, linktype=None, extension=None):
    """Return a link to the result.
    Linktype can be specified as parameter or as a state variable (e.g. Set~linktype~query)
    linktype can be
    - query : just a query from the state - without the last command (Link) if removelast is True
    - dataurl : data URL (base64-encoded)
    - htmlimage : html image element
    - path : absolute url path (without server)
    - url : complete url
    """
    q = state.query
    if linktype in (None, "default"):
        linktype = state.vars.get("linktype", "query")
    if linktype == "query":
        return state.with_data(q)
    elif linktype == "dataurl":
        b, mime, typeid = encode_state_data(state.get(), extension=extension)
        encoded = base64.b64encode(b).decode("ascii")
        return state.with_data(f"data:{mime};base64,{encoded}")
    elif linktype == "htmlimage":
        b, mime, typeid = encode_state_data(state.get(), extension=extension)
        encoded = base64.b64encode(b).decode("ascii")
        return state.with_filename("image.html").with_data(
            f'<img src="data:{mime};base64,{encoded}"/>'
        )
    elif linktype == "path":
        return state.with_data(state.vars.get("api_path", "/q/") + q)
    elif linktype == "url":
        return state.with_data(
            state.vars.get("server", "http://localhost")
            + state.vars.get("api_path", "/q/")
            + q
        )

    raise Exception(f"Unsupported link type: {linktype}")


@command
def html_a(state, linktype=None, value=None, extension=None):
    """Create a html link tag
    If value is None, query is displayed as a value.
    """
    if value is None:
        value = state.query

    lnk = link(state, linktype=linktype, extension=extension)
    return state.with_data(f'<a href="{lnk}">{value}</a>')


@first_command
def fetch(url):
    """Load binary data from URL"""
    return urlopen(url).read()


@first_command
def fetch_utf8(url):
    """Load text data encoded as utf-8 from URL"""
    return urlopen(url).read().decode("utf-8")


@command
def filename(state, name):
    """Set filename"""
    return state.with_filename(name)


@command
def ns(state, *namespaces):
    """Set command namespaces to be included

    By default only the root namespace is enabled. The 'ns' command enables commands present in specified namespaces.
    This works by setting "active_namespaces" state variable.
    The "root" namespace is appended to the active_namespaces if not already present.
    When command is executed, all the active namespaces are searched one by one until the command is found.
    Note that order of active_namespaces is significant.
    """
    namespaces = list(namespaces)
    if "root" not in namespaces:
        namespaces.append("root")
    state.vars["active_namespaces"] = namespaces

    return state


@first_command(volatile=True)
def clean_cache():
    print(f"clean cache {get_cache()}")
    get_cache().clean()
    return "Cache cleaned"


@first_command(volatile=True)
def queries_status(include_ready=False, reduce=True):
    import liquer.parser as lp
    import traceback

    try:
        cache = get_cache()
        data = []
        for key in sorted(cache.keys()):
            metadata = cache.get_metadata(key)
            if metadata is None:
                continue
            progress = metadata.get("progress_indicators", []) + metadata.get(
                "child_progress_indicators", []
            )
            d = dict(
                query=key,
                short=lp.parse(key).short(),
                status=metadata.get("status", "none"),
                updated=metadata.get("updated", "?"),
                message=metadata.get("message", ""),
                progress=progress[:3],
            )
            if include_ready or d["status"] != Status.READY.value:
                data.append((key, d))
        data = [x[1] for x in sorted(data)]
        if reduce and len(data):
            reduced_data = [data[0]]
            for d, next_d in zip(data[1:], data[2:]):
                previous_d = reduced_data[-1]
                if not (
                    previous_d["status"] == Status.EVALUATING_PARENT.value
                    and d["status"] == Status.EVALUATING_PARENT.value
                    and d["query"].startswith(previous_d["query"])
                    and next_d["query"].startswith(d["query"])
                ):
                    reduced_data.append(d)
            reduced_data.append(data[-1])
            data = reduced_data
        return data
    except:
        return [
            dict(
                query="",
                status="not available",
                updated="",
                message=traceback.format_exc(),
                progress=[],
            )
        ]


@command
def dr(state, type_identifier=None, extension=None, context=None):
    """Decode resource
    Decodes the bytes into a data structure. This is meant to be used in connection to a resource query.
    Resource part of the query will typically fetch the data from a store and thus return bytes (together with metadata).
    Command dr will convert the bytes (assuming proper metadata are provided) into a data structure.
    The metadata must contain type_identifier in metadata or metadata['resource_metadata'], a filename with extension
    or extension with known decoding.
    """
    from liquer.state_types import state_types_registry
    from liquer.parser import parse

    if state.data is None:
        context.error(
            f"Bytes expected, None received in dr from {state.query}")
        return

    if type_identifier is None:
        type_identifier = state.metadata.get(
            "type_identifier",
            state.metadata.get("resource_metadata", {}).get("type_identifier"),
        )

    if type_identifier in (None, "bytes"):
        type_identifier = state.metadata.get(
            "resource_metadata", {}).get("type_identifier")

    if extension is None:
        extension = state.metadata.get("extension")
    if extension is None:
        query = state.metadata.get("query")
        if query is not None:
            filename = parse(query).filename()
        if filename is not None:
            v = filename.split(".")
            if len(v) > 1:
                extension = v[-1]
                context.info(f"Extension: {extension} - from query '{query}'")
        else:
            key = state.metadata.get("resource_metadata", {}).get("key")
            if key is not None:
                filename = context.store().key_name(key)
            v = filename.split(".")
            if len(v) > 1:
                extension = v[-1]
                context.info(f"Extension: {extension} - from key '{key}'")

    if type_identifier in (None, "bytes"):
        type_identifier = type_identifier_from_extension(extension)
        context.info(
            f"Type identifier: {type_identifier} - from extension '{extension}'")

    if type_identifier is not None:
        if extension in ("parquet", "xlsx", "csv", "tsv") and type_identifier in ("generic", "dictionary", "pickle"):
            context.warning(f"Type identifier '{type_identifier}' seems to be inconsistent with the extension '{extension}'")
            context.warning(f"This might indicate a problem with executing the partent query '{context.parent_query}'")
            type_identifier = type_identifier_by_extension.get(extension)
            context.warning(
                f"To fix the inconsistency, type identifier: {type_identifier} is used from extension '{extension}'")
            
        context.info(
            f"Type identifier: {type_identifier},  Extension: {extension}")
        t = state_types_registry().get(type_identifier)
        data = t.from_bytes(state.data, extension=extension)
        return state.with_data(data)
    else:
        context.error(f"Decode resource (dr) command failed")
        raise Exception(f"Failed to resolve type for query {state.metadata.get('query')}")
    return state

@first_command
def sync_store(context=None):
    context = get_context(context)
    context.info(f"Sync store")
    context.store().sync()
    return dict(status="OK", message="Store synchronized")



import multiprocessing as mp
from multiprocessing.managers import BaseManager
from os import getpid
from time import sleep
from liquer.cache import set_cache, CacheProxy, NoCache
from liquer import *
from liquer.constants import *

class PoolManager(BaseManager):
    pass


PoolManager.register("CacheProxy", CacheProxy)

_pool = None


def start_pool(processes=None):
    global _pool
    _pool = mp.Pool(processes=processes)
    return _pool


def get_pool():
    global _pool
    if mp.current_process().name != "MainProcess":
        raise Exception("Pool should only be accessed from the main process")
    if _pool is None:
        start_pool()
    return _pool


_worker_config = None


def get_cache(worker_config):
    print(f"get_cache({worker_config})")
    if worker_config is None:
        print("worker config missing => No cache")
        return NoCache()
    if "cache" in worker_config:
        cache = worker_config["cache"]
        print(f"Central cache {repr(cache)}")
        return cache
    elif "cache_constructor" in worker_config:
        f = worker_config["cache_constructor"]
        arg = worker_config.get("cache_arg", [])
        if arg is None:
            arg = []
        kwarg = worker_config.get("cache_kwarg", {})
        if kwarg is None:
            kwarg = {}
        cache = f(*arg, **kwarg)
        print(f"Local cache {repr(cache)}")
        return cache
    else:
        print(f"Worker confing missging cache description: {worker_config}")
    return NoCache()


def set_local_cache_constructor(constructor, arg=None, kwarg=None):
    """Set the pool cache to use a locally constructed cache.
    This means that each pool worker will construct its cache locally
    applying arguments arg and keyword arguments kwarg to the constructor.
    This is suitable for FileCache (more or less) or a server-based cache (e.g. SQLCache with a database server),
    but it is not suitable for MemoryCache, since each worker will have its own memory cache and thus the cache will not be shared.

    In general this is a less secure solution. Even for FileCache there might be collisions if multiple workers
    try to access the same file. The set_central_cache should be a safe alternative.
    """
    global _worker_config
    if _worker_config is None:
        _worker_config = {}
    _worker_config.update(
        dict(cache_constructor=constructor, cache_arg=arg, cache_kwarg=kwarg)
    )
    cache = constructor(*arg, **kwarg)
    set_cache(cache)


def set_central_cache(cache, manager=None, use_cache_proxy_locally=True):
    """Set the pool cache to use a cache object.
    The cache object lives in a single process, which is then accessed from workers using IPC.
    This is suitable for all types of cache objects, particularly it is important for a MemoryCache.

    In general this is a more secure solution, but it requires all data to support pickle.
    It has to be noted that the cache lives in a single process and relies on IPC to work, thus this solution
    may possibly be outperformed by set_local_cache_constructor.

    PoolManager instance can be provided via manager. If None, PollManager will be created and started.
    The manager instance is returned.

    The use_cache_proxy_locally controls if cache is proxied when used locally (which should be the safe choice).
    """
    global _worker_config
    if manager is None:
        manager = PoolManager()
        manager.start()
    if _worker_config is None:
        _worker_config = {}
    cache_proxy = manager.CacheProxy(cache)
    _worker_config.update(dict(cache=cache_proxy))
    if use_cache_proxy_locally:
        set_cache(cache_proxy)
    else:
        set_cache(cache)
    return manager


def _evaluate_worker(query, worker_config):
    print(f"Evaluate worker started for {query}")
    set_cache(get_cache(worker_config))
    evaluate(query)
    return f"Done evaluating {query}"


def _evaluate_and_save_worker(query, target_directory, target_file, worker_config):
    print(f"Evaluate and save worker started for {query}")
    set_cache(get_cache(worker_config))
    evaluate_and_save(query, target_directory=target_directory, target_file=target_file)
    return f"Done evaluate and save {query}"


def evaluate_in_background(query):
    """Like evaluate, but returns immediately and runs in the background.
    Note that this creates immediately a submitted state in the cache.

    If there is no metadata associated with the query after calling this function,
    then either there is no cache or the resulting state is volatile.
    This can be used by GUI to identify situations when waiting for the result to appear
    in the cache is not a viable strategy, but requesting the object directly is necessary.
    """
    global _worker_config
    print(f"Evaluate {query} in background")
    metadata = get_context().metadata()
    metadata["query"]=query
    metadata["status"]=Status.SUBMITTED.value
    cache = get_cache(_worker_config)
    cache.store_metadata(metadata)

    if _worker_config is None:
        print("WARNING: Evaluated in main process")
        print(
            "         Please configure the cache using set_local_cache_constructor or set_central_cache"
        )
        return evaluate(query)
    else:
        return get_pool().apply_async(
            _evaluate_worker, [query, _worker_config], callback=print
        )


def evaluate_and_save_in_background(query, target_directory=None, target_file=None):
    """Like evaluate_and_save, but returns immediately and runs in the background.
    Note that the saving occurs on o worker. Eventual remote workers will save the results in their local filesystem.
    """
    global _worker_config
    print(f"Evaluate and save {query} in background")
    if _worker_config is None:
        print("WARNING: Evaluated in main process")
        print(
            "         Please configure the cache using set_local_cache_constructor or set_central_cache"
        )
        return evaluate_and_save(
            query, target_directory=target_directory, target_file=target_file
        )
    return get_pool().apply_async(
        _evaluate_and_save_worker,
        [query, target_directory, target_file, _worker_config],
        callback=print,
    )



import traceback
from liquer.state import State, EvaluationException, vars_clone
from liquer.parser import (
    encode,
    decode,
    parse,
    QueryException,
    TransformQuerySegment,
    Query,
    ActionRequest,
    StringActionParameter,
    ExpandedActionParameter,
    LinkActionParameter,
)
from liquer.cache import cached_part, get_cache
from liquer.commands import command_registry
from liquer.state_types import (
    encode_state_data,
    state_types_registry,
    data_characteristics,
    type_identifier_of,
)
import os.path
from datetime import datetime
import json
from liquer.constants import Status, mimetype_from_extension
import liquer.util as util
from liquer.util import timestamp
from copy import deepcopy
from liquer.metadata import Metadata

from liquer.store import get_store, Store, KeyNotFoundStoreException, StoreException, key_extension


def find_queries_in_template(template: str, prefix: str, sufix: str):
    try:
        start = template.index(prefix)
        end = template.index(sufix, start + len(prefix))
        yield template[:start], template[start + len(prefix) : end]
        for text, query in find_queries_in_template(
            template[end + len(sufix) :], prefix, sufix
        ):
            yield text, query
    except ValueError:
        yield template, None


class Vars(dict):
    def __init__(self, *arg, **kwarg):
        super().__init__(*arg, **kwarg)
        self._modified_vars = set()

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        if name == "_modified_vars":
            super().__setattr__(name, value)
        else:
            self._modified_vars.add(name)
            self[name] = value

    def __getstate__(self):
        return (self._modified_vars, dict(self))

    def __setstate__(self, x):
        mv, d = x
        super().__setattr__("_modified_vars", mv)
        self.clear()
        self.update(d)

    def get_modified(self):
        return {key: self[key] for key in self._modified_vars}


def log_time():
    date = datetime.now()
    return date.strftime("%H:%M:%S")


CONTEXT_CREATOR = None


def set_context_creator(context_creator):
    global CONTEXT_CREATOR
    CONTEXT_CREATOR = context_creator


def get_context(context=None):
    global CONTEXT_CREATOR
    if context is None:
        if CONTEXT_CREATOR is None:
            return Context()
        else:
            return CONTEXT_CREATOR()
    else:
        return context

class MetadataContextMixin:
    def metadata(self):
        metadata = self._metadata.as_dict()
        title = self.title
        description = self.description
        if title is None:
            if self.raw_query is None:
                title = ""
            else:
                p = parse(self.raw_query)
                if title in ("", None):
                    title = p.filename() or ""

        mimetype = self.mimetype
        if mimetype is None:
            if self.query is not None:
                if self.query.extension() is None:
                    mimetype = "application/octet-stream"
                else:
                    mimetype = mimetype_from_extension(self.query.extension())

        message = self._metadata.message
        if message in (None, ""):
            log = self._metadata.get("log",[])
            if len(log):
                message = log[-1]["message"]
        if message in (None, ""):
            log = self._metadata.get("child_log",[])
            if len(log):
                message = log[-1]["message"]

        metadata.update(dict(
            status=self.status.value,
            title=title,
            description=description,
            mimetype=mimetype,
            query=self.raw_query,
            parent_query=self.parent_query,
            argument_queries=self.argument_queries,
#            log=self.log[:],
            is_error=self.is_error,
            direct_subqueries=self.direct_subqueries[:],
            progress_indicators=self.progress_indicators[:],
            child_progress_indicators=self.child_progress_indicators[:],
            child_log=self.child_log,
            message=message,
            started=self.started,
            updated=self.now(),
            created=self.created,
            caching=self.caching,
            vars=dict(self.vars),
            html_preview=self.html_preview,
            side_effect=False,
        ))
        return metadata

    def log_dict(self, d):
        "Put dictionary with a log entry into the log"
        d["timestamp"] = timestamp()
        self._metadata.log_dict(d)
        self.store_metadata(force=(d.get("kind") == "error"))
        if self.parent_context is not None:
            if d.get("origin") is None:
                d = dict(origin=self.raw_query, **d)
            self.parent_context.log_child_dict(d)
        return self

    def log_action(self, qv, number=0):
        """Log a command"""
        if isinstance(qv, ActionRequest):
            qv = qv.to_list()
        return self.log_dict(dict(kind="command", qv=qv, command_number=number))

    def error(self, message, position=None, query=None):
        """Log an error message"""
        self.is_error = True
        self.status = Status.ERROR
        if position is None:
            print(f"{log_time()} ERROR:    ", message)
        else:
            print(f"{log_time()} ERROR:    ", message, f" at {position}")
        return self.log_dict(
            dict(
                kind="error",
                message=message,
                position=None if position is None else position.to_dict(),
                query=query,
            )
        )

    def warning(self, message, traceback=None):
        """Log a warning message"""
        print(f"{log_time()} WARNING:  ", message)
        return self.log_dict(dict(kind="warning", message=message, traceback=traceback))

    def exception(self, message, traceback, position=None, query=None):
        """Log an exception"""
        self.is_error = True
        self.status = Status.ERROR
        if position is None:
            print(f"{log_time()} EXCEPTION:", message)
        else:
            print(f"{log_time()} EXCEPTION:", message, f" at {position}")
        return self.log_dict(
            dict(
                kind="error",
                message=message,
                traceback=traceback,
                position=None if position is None else position.to_dict(),
                query=query,
            )
        )

    def info(self, message):
        """Log a message (info)"""
        print(f"{log_time()} INFO:     ", message)
        self.log_dict(dict(kind="info", message=message))
        return self

    def debug(self, message):
        """Log a message (info)"""
        if self.debug_messages:
            print(f"{log_time()} DEBUG:    ", message)
            self.log_dict(dict(kind="debug", message=message))
        return self

    @property
    def raw_query(self):
        return self._metadata.query

    @raw_query.setter
    def raw_query(self, value):
        self._metadata.query = value

class Context(MetadataContextMixin, object):
    def __init__(self, parent_context=None, debug=False):
        self.parent_context = parent_context  # parent context - when in child context

#        self.raw_query = None  # String with the evaluated query
        self.query = None  # Query object of the evaluated query
        self.status = Status.NONE  # Status: ready, error...
        self.is_error = False  # True is evaluation failed

        self.started = ""  # Evaluation start time
        self.created = ""  # Created time (evaluation finished)

        self.direct_subqueries = (
            []
        )  # list of subqueries specified as dictionaries with description and query
        self.parent_query = None  # parent query or None
        self.argument_queries = (
            []
        )  # list of argument subqueries specified as dictionaries with description and query

        self.progress_indicators = []  # progress indicators as a list of dictionaries
        #self.log = []  # log of messages as a list of dictionaries
        self.child_progress_indicators = []  # progress indicator of a child
        self.child_log = []  # log of messages from child queries
        self.message = ""  # Last message from the log
        self.debug_messages = debug  # Turn the debug messages on/off
        self.caching = True  # caching of the results enabled
        self.enable_store_metadata = True  # flag to controll storing of metadata

        self.last_report_time = None  # internal time stamp of the last report
        self._progress_indicator_identifier = (
            1  # counter for creating unique progress identifiers
        )
        self.description = ""
        self.title = None
        self.mimetype = None

        self.vars = Vars(vars_clone())
        self.html_preview = ""
        self.store_key = None
        self.store_to = None

        self._metadata = Metadata()

    def new_empty(self):
        return Context(debug=self.debug_messages)


    def store_data(self, key, data):
        """Convenience method to store data in the store including metadata.
        Note that the metadata are taken from the context.metadata() and slightly updated.
        They might not be 100% correct, since the store_data will probably be called as
        a side-effect of a query, not a result of a query.
        This is indicated by the side_effect flag in the metadata and status Status.SIDE_EFFECT.value.
        """
        metadata = self.metadata()
        store = self.store()
        extension = key_extension(key)
        metadata["data_characteristics"] = data_characteristics(data)
        metadata["side_effect"] = True
        metadata["status"] = Status.SIDE_EFFECT.value
        try:
            b, mimetype, type_identifier = encode_state_data(data, extension=extension)
            metadata["type_identifier"] = type_identifier
            metadata["mimetype"] = mimetype
            store.store(key, b, metadata)
        except:
            traceback.print_exc()
            m = Metadata(metadata)
            m.status = Status.ERROR
            m.exception(
                f"Failed to encode data for key '{key}'",
                traceback=traceback.format_exc(),
            )
            store.store_metadata(key, m.as_dict())

    def can_report(self):
        if self.last_report_time is None:
            self.last_report_time = datetime.now()
        return True
        return (datetime.now() - self.last_report_time).total_seconds() > 0.1

    def set_html_preview(self, html):
        self.html_preview = html
        self.store_metadata()
        return self

    def set_description(self, description):
        self.description = description
        self.store_metadata()
        return self

    def set_title(self, title):
        self.title = title
        self.store_metadata()
        return self

    def enable_cache(self, enable=True):
        self.caching = enable
        return self

    def disable_cache(self):
        self.enable_cache(False)
        return self

    def create_state(self):
        return State(metadata=self.metadata(), context=self)

    def store_metadata(self, force=False):
        if self.raw_query is not None and self.enable_store_metadata:
            if force or self.can_report():
                metadata = self.metadata()
                self.cache().store_metadata(self.metadata())
                self.last_report_time = datetime.now()
                if self.store_key is not None:
                    store = self.store() if self.store_to is None else self.store_to
                    store.store_metadata(self.store_key, metadata)

    def new_progress_indicator(self):
        self._progress_indicator_identifier += 1
        self.progress_indicators.append(
            dict(
                id=self._progress_indicator_identifier,
                step=0,
                total_steps=None,
                message="",
                timestamp=timestamp(),
            )
        )
        return self._progress_indicator_identifier

    def remove_progress_indicator(self, identifier):
        self.progress_indicators = [
            x for x in self.progress_indicators if x["id"] != identifier
        ]
        if self.parent_context is not None:
            self.parent_context.remove_child_progress(self.raw_query)

    def progress_indicator_index(self, identifier):
        if identifier is None:
            if len(self.progress_indicators):
                return len(self.progress_indicators) - 1
            self.new_progress_indicator()
            return len(self.progress_indicators) - 1

        for i, x in enumerate(self.progress_indicators):
            if x["id"] == identifier:
                return i
        return None

    def now(self):
        return util.now()

    def progress(
        self, step=0, total_steps=None, message="", identifier=None, autoremove=True
    ):
        index = self.progress_indicator_index(identifier)

        progress = dict(
            step=step, total_steps=total_steps, message=message, timestamp=timestamp()
        )
        self.progress_indicators[index].update(progress)

        removed = False
        if autoremove and total_steps is not None and step >= total_steps:
            self.remove_progress_indicator(index)
            removed = True

        self.store_metadata()

        if self.parent_context is not None:
            if removed:
                self.parent_context.remove_child_progress(self.raw_query)
            else:
                d = dict(origin=self.raw_query, **progress)
                self.parent_context.log_child_progress(d)

    def progress_iter(self, iterator, show_value=False):
        try:
            total_steps = len(iterator)
        except:
            total_steps = None
        identifier = self.new_progress_indicator()
        for i, x in enumerate(iterator):
            if total_steps is None:
                message = f"{x} ({i+1})" if show_value else f"{i+1}"
            else:
                message = (
                    f"{x} ({i+1}/{total_steps})"
                    if show_value
                    else f"{i+1}/{total_steps}"
                )
            self.progress(
                i, total_steps=total_steps, message=message, identifier=identifier
            )
            yield x
        self.remove_progress_indicator(identifier)

    def remove_child_progress(self, origin):
        "Remove all child progress indicators from a given origin"
        self.child_progress_indicators = [
            x for x in self.child_progress_indicators if x.get("origin") != origin
        ]
        self.store_metadata()
        if self.parent_context is not None:
            self.parent_context.remove_child_progress(origin)

    def log_child_progress(self, d):
        "Put dictionary with a child progress entry into the child progress indicators and notify parent"
        self.child_progress_indicators = [
            x
            for x in self.child_progress_indicators
            if x.get("origin") != d.get("origin")
        ]
        self.child_progress_indicators.append(d)
        self.store_metadata()
        if self.parent_context is not None:
            self.parent_context.log_child_progress(d)
        return self

    def log_child_dict(self, d):
        "Put dictionary with a child log entry into the child log"
        d = dict(**d)
        if d.get("origin") is None:
            if self.parent_context is None:
                d["origin"] = None
            else:
                d["origin"] = self.parent_context.raw_query
        self.child_log.append(d)
        self.child_log = self.child_log[:5]
        self.store_metadata()
        if self.parent_context is not None:
            self.parent_context.log_child_dict(d)
        return self


    def child_context(self):
        return self.__class__(parent_context=self)

    def root_context(self):
        return (
            self if self.parent_context is None else self.parent_context.root_context()
        )

    def log_subquery(self, query: str, description=None):
        assert type(query) == str
        if query not in self.direct_subqueries:
            if description is None:
                description = query
            self.direct_subqueries.append(dict(description=description, query=query))

    def command_registry(self):
        return command_registry()

    def cache(self):
        return get_cache()

    def state_types_registry(self):
        return state_types_registry()

    def evaluate_parameter(self, p, action):
        if isinstance(p, StringActionParameter):
            return p
        elif isinstance(p, LinkActionParameter):
            if p.link.absolute:
                self.argument_queries.append(
                    dict(
                        description=f"{p.link.encode()} at {p.position}",
                        query=p.link.encode(),
                    )
                )
                self.debug(f"Expand absolute link parameter {p.link.encode()}")
                value = self.evaluate(p.link)
                if value.is_error:
                    self.error(
                        f"Error while evaluating absolute link parameter {p.link.encode()}",
                        position=p.position,
                        query=self.raw_query,
                    )
                    self.status = Status.ERROR
                    self.store_metadata(force=True)

                    raise EvaluationException(
                        f"Error while evaluating absolute link parameter {p.link.encode()} in action {action.name}",
                        position=p.position,
                        query=self.raw_query,
                    )
                pp = ExpandedActionParameter(value.get(), p.link, p.position)
                return pp
            else:
                self.argument_queries.append(
                    dict(
                        description=f"{p.link.encode()} at {p.position}",
                        query=p.link.encode(),
                    )
                )
                self.debug(
                    f"Expand relative link parameter {p.link.encode()} on {self.parent_query}"
                )
                value = self.apply(p.link)
                if value.is_error:
                    self.error(
                        f"Error while evaluating relative link parameter {p.link.encode()} at {p.position}",
                        position=p.position,
                        query=self.raw_query,
                    )
                    self.status = Status.ERROR
                    self.store_metadata(force=True)

                    raise EvaluationException(
                        f"Error while evaluating relative link parameter {p.link.encode()} in action {action.name}",
                        position=p.position,
                        query=self.raw_query,
                    )
                pp = ExpandedActionParameter(value.get(), p.link, p.position)
                return pp
        else:
            self.status = Status.ERROR
            self.store_metadata(force=True)
            raise EvaluationException(
                f"Unknown parameter type {type(p)} in {action.name}",
                position=action.position,
                query=self.raw_query,
            )

    def evaluate_action(self, state: State, action, cache=None):
        self.debug(f"EVALUATE ACTION '{action}' on '{state.query}'")
        self.status = Status.EVALUATION
        self.store_metadata(force=True)
        cache = cache or self.cache()
        cr = self.command_registry()

        state.context = self

        if isinstance(action, TransformQuerySegment):
            if action.is_filename():
                return state.with_filename(action.filename)
            assert action.is_action_request()
            action = action.query[0]

        is_volatile = state.is_volatile()
        old_state = state if is_volatile else state.clone()

        state = state.next_state()
        state.context = self

        ns, command, cmd_metadata = cr.resolve_command(state, action.name)
        if command is None:
            self.error(
                f"Unknown action: '{action.name}'",
                position=action.position,
                query=self.raw_query,
            )
        else:
            self._metadata.add_command_dependency(ns, cmd_metadata)
            parameters = []
            self.status = Status.EVALUATING_DEPENDENCIES
            self.store_metadata(force=True)
            for p in action.parameters:
                parameters.append(self.evaluate_parameter(p, action))
            self.status = Status.EVALUATION
            self.store_metadata(force=True)

            try:

                state = command(old_state, *parameters, context=self)
                assert type(state.metadata) is dict
            except EvaluationException as ee:
                print("EE:", ee)
                # traceback.print_exc()
                state.is_error = True
                state.exception = ee
            except Exception as e:
                traceback.print_exc()
                state.is_error = True
                self.exception(
                    message=str(e),
                    position=action.position,
                    query=self.raw_query,
                    traceback=traceback.format_exc(),
                )
                state.exception = EvaluationException(
                    traceback.format_exc() + "\n" + str(e),
                    position=action.position,
                    query=self.raw_query,
                )
        arguments = getattr(state, "arguments", None)
        if arguments is not None:

            def to_arg(arg):
                x, meta = arg
                try:
                    s = json.dumps(x)
                    if len(s) > 100:
                        return [s[:50], meta]
                    return [x, meta]
                except:
                    return [None, meta]

            arguments = [to_arg(a) for a in arguments]

        metadata = self.metadata()
        metadata["type_identifier"]=state.type_identifier
        metadata["commands"] = metadata.get("commands", []) + [action.to_list()]
        if metadata.get("mimetype","application/octet-stream") == "application/octet-stream":
            metadata["mimetype"] = state.mimetype()
            
        try:
            cmd_metadata_d = cmd_metadata._asdict()
        except:
            cmd_metadata_d = {}
        metadata["extended_commands"] = metadata.get("extended_commands", []) + [
            dict(
                command_name=action.name,
                ns=ns,
                qcommand=action.to_list(),
                action=f"{action.encode()} at {action.position}",
                command_metadata=cmd_metadata_d,
                arguments=arguments,
            )
        ]
        metadata["query"] = self.raw_query
        metadata["attributes"] = {
            key: value
            for key, value in state.metadata["attributes"].items()
            if key[0].isupper()
        }

        if cmd_metadata is not None:
            metadata["attributes"] = dict(
                metadata.get("attributes", {}), **cmd_metadata.attributes
            )

        metadata["caching"] = metadata.get("caching", True) and state.metadata.get(
            "caching", True
        )
        is_error = state.is_error or self.is_error

        if is_error:
            self.status = Status.ERROR
            metadata["status"] = self.status.value
            self.info(f"Action {action.encode()} at {action.position} failed")
            state.metadata.update(metadata)
            state.status = Status.ERROR.value
            state.is_error = True
        else:
            self.status = Status.READY
            metadata["status"] = self.status.value
            self.info(f"Action {action.encode()} at {action.position} completed")
            state_vars = dict(self.vars)
            state_vars.update(state.vars)
            state_vars.update(self.vars.get_modified())
            self.vars = Vars(state_vars)
            metadata["vars"] = dict(state_vars)
            state.metadata.update(metadata)

        state.set_volatile(is_volatile or state.is_volatile())

        cache.store_metadata(state.metadata)
        return state

    def store(self):
        return get_store()

    def evaluate_resource(self, resource_query):
        self.info(f"Evaluate resource: {resource_query}")
        if resource_query.header is not None:
            if resource_query.header.encode() not in ("-R", "-R-meta"):
                raise Exception(
                    f"Header '{resource_query.header}' not supported in resource query {resource_query}"
                )
        key = resource_query.path()
        store = self.store()
        state = self.create_initial_state()
        try:
            metadata = store.get_metadata(key)
            if metadata is None:
                if store.contains(key):
                    state.error(
                        f"Key '{key}' was found in store, but the metadata is missing."
                    )
                else:
                    state.error(f"Metadata for key '{key}' not found in store")
            if (
                resource_query.header is not None
                and len(resource_query.header.parameters) > 0
                and resource_query.header.parameters[-1].encode() == "meta"
            ):
                self.info(f"Resource metadata query {resource_query}")
                data = metadata
                metadata = dict(
                    description=f"Metadata for {key}",
                    key=key,
                    query=resource_query.encode(),
                )
            else:
                data = store.get_bytes(key)
                if data is None:
                    if store.contains(key):
                        if store.is_dir(key):
                            state.error(
                                f"Key '{key}' is a directory, hence there is no data."
                            )
                        else:
                            state.error(
                                f"Key '{key}' was found in store, but the data is missing."
                            )
                    else:
                        state.error(f"Key '{key}' not found in store")

            state = state.with_data(data)
            state.metadata["resource_metadata"] = metadata
        except:
            self.exception(
                message=f"Error evaluating resource {resource_query}",
                traceback=traceback.format_exc(),
                position=resource_query.position,
                query=resource_query.encode(),
            )
            traceback.print_exc()
        return state

    def create_initial_state(self):
        state = State()
        state.query = ""
        return state

    @classmethod
    def to_query(cls, query):
        if query is None:
            return "", Query()
        if isinstance(query, str):
            return query, parse(query)
        elif isinstance(query, Query):
            return query.encode(), query
        else:
            raise Exception(f"Unsupported query type: {type(query)}")

    def apply(self, query, description=None):
        self.debug(f"APPLY {query}")
        if self.parent_query in (None, "", "/"):
            self.debug(f"  no parent query in apply {query}")
            return self.evaluate(query, description=description)
        if isinstance(query, str):
            query = parse(query)
        if query.absolute:
            self.debug(f"  absolute link in apply {query}")
            return self.evaluate(query, description=description)
        tq = query.transform_query()
        if tq is None:
            raise Exception(
                f"Only transform query supported in apply ({query} on {self.parent_query})"
            )
        q = (parse(self.parent_query) + tq).encode()
        self.debug(f"apply {query} on {self.parent_query} yields {q}")
        return self.evaluate(q, description=description)

    def _store_state(self, state):
        if self.store_key is not None:
            metadata = state.metadata
            store = self.store() if self.store_to is None else self.store_to
            if state.is_error:
                store.store_metadata(self.store_key, metadata)
            else:
                data = state.get()
                reg = self.state_types_registry()
                t = reg.get(type(data))
                try:
                    if state.metadata.get("extension") is None:
                        b, mime, typeid = encode_state_data(data)
                    else:
                        b, mime, typeid = encode_state_data(
                            data, extension=state.metadata["extension"]
                        )
                    store.store(self.store_key, b, metadata)
                except:
                    traceback.print_exc()
                    m = Metadata(metadata)
                    m.status = Status.ERROR
                    m.exception(
                        f"Failed to encode data for key '{self.store_key}'",
                        traceback=traceback.format_exc(),
                    )
                    store.store_metadata(self.store_key, m.as_dict())

    def evaluate(
        self, query, cache=None, description=None, store_key=None, store_to=None
    ):
        """Evaluate query, returns a State.
        This method can be used in a command to evaluate a subquery,
        which will be recorded in metadata and can be inspected during the query execution.

        When evaluating such a subquery, it is good to give it a description (via a description argument).
        If this is not a sub-query, description parameter will set the description in the metadata (see set_description).
        Note that this might be overridden by the subsequent calls to set_description.
        Parameter cache can be used to set a cache object. In most cases default cache is the safest choice.

        Evaluation can be (besides cache) stored in the store under the key specified by the store_key (if not None).
        A store can be specified too via the store_to option. If None (default), the default store (from the store method) is used.
        """
        self.enable_store_metadata = False  # Prevents overwriting cache with metadata
        self.status = Status.EVALUATION
        self.debug(f"EVALUATE {query} ")

        self.vars = Vars(vars_clone())

        if self.query is not None:
            self.enable_store_metadata = True
            print(f"Subquery {query} called from {self.query.encode()}")
            state = self.child_context().evaluate(
                query, store_key=store_key, store_to=store_to
            )
            if not isinstance(query, str):
                query = query.encode()
            self.log_subquery(query=query, description=description)
            if state.is_error:
                print("Subquery failed")
                for d in state.metadata.get("log", []):
                    self.log_dict(d)
            #            self.enable_store_metadata = True
            self.store_metadata(force=True)
            self.enable_store_metadata = False
            return state

        raw_query, query = self.to_query(query)
        self.raw_query=raw_query
        self.query = query
        self.store_key = store_key
        self.store_to = store_to
        self.started = self.now()
        if description is not None:
            self.set_description(description)

        if cache is None:
            cache = self.cache()

        self.debug(f"Using cache {repr(cache)}")
        self.debug(f"Try cache {query}")
        state = cache.get(query.encode())
        if state is not None:
            self.debug(f"Cache hit {query}")
            self._store_state(state)
            return state
        self.enable_store_metadata = (
            True  # Metadata can be only written after trying to read from cache,
        )
        # so that cache does not get overwritten
        self.debug(f"Cache miss {query}")

        if query.is_resource_query():
            state = self.evaluate_resource(query.resource_query())
            state.query = query.encode()
            state.metadata["created"] = self.now()
            self._store_state(state)
            return state
        else:
            p, r = query.predecessor()
            self.debug(f"PROCESS Predecessor:{p} Action: {r}")
            if p is None or p.is_empty():
                self.parent_query = ""
                state = self.create_initial_state()
                state.metadata["created"] = self.now()
                self.debug(f"INITIAL STATE")
            else:
                self.parent_query = p.encode()
                self.status = Status.EVALUATING_PARENT
                self.store_metadata(force=True)
                state = self.child_context().evaluate(p, cache=cache)
            if state.is_error:
                self.status = Status.ERROR
                self.store_metadata()
                state = state.next_state()
                state.query = query.encode()
                state.metadata["created"] = self.now()
                self.debug(f"ERROR in '{state.query}'")
                self._store_state(state)
                return state
        self.vars = Vars(state.vars)
        if r is None:
            self.debug(f"RETURN '{query}' AFTER EMPTY ACTION ON '{state.query}'")
            state.query = query.encode()
            state.metadata["created"] = self.now()
            self._store_state(state)
            return state
        elif r.is_filename():
            state.metadata["filename"] = r.filename
            state.metadata["extension"] = ".".join(r.filename.split(".")[1:])

        state = self.evaluate_action(state, r)
        state.query = query.encode()
        state.metadata["created"] = self.now()

        if (
            state.metadata.get("caching", True)
            and not state.is_error
            and not state.is_volatile()
        ):
            print("CACHE", state.query)
            #            self.status = "cache"
            #            self.store_metadata()
            cache.store(state)
        else:
            if state.is_error:
                cache.store_metadata(state.metadata)
            else:
                print("REMOVE CACHE", state.query)
                if not cache.remove(state.query):
                    self.status = Status.EXPIRED
                    self.store_metadata()

        self._store_state(state)
        return state

    def evaluate_and_save(
        self,
        query,
        target_directory=None,
        target_file=None,
        target_resource_directory=None,
        store=None,
    ):
        """Evaluate query and save result.
        Output is saved either to
        - a target directory (current working directory by default) to a file deduced from the query, or
        - to target_file (if specified)
        Returns a state.
        """

        if target_directory==None and target_file==None and target_resource_directory==None and store==None:
            target_directory = "."

        print(f"*** Evaluate and save {query} started")
        state = self.evaluate(query)
        if state.is_error:
            print(f"*** Evaluate and save {query} failed")
            if target_resource_directory is not None and target_file is not None:
                filename = target_file
                key = (
                    filename
                    if target_resource_directory == ""
                    else target_resource_directory + "/" + filename
                )
                if store is None:
                    store = self.store()

                store.store_metadata(key, state.metadata)

        data = state.get()
        reg = self.state_types_registry()
        t = reg.get(type(data))

        filename = target_file
        if state.metadata.get("extension") is None:
            b, mime, typeid = encode_state_data(data)
            filename = t.default_filename() if target_file is None else target_file
        else:
            b, mime, typeid = encode_state_data(
                data, extension=state.metadata["extension"]
            )
            filename = (
                (
                    t.default_filename()
                    if state.metadata.get("filename") is None
                    else state.metadata["filename"]
                )
                if target_file is None
                else target_file
            )
        if target_directory is None:
            path = filename
        else:
            path = os.path.join(target_directory, filename)

        if target_directory is not None:
            print(f"*** Evaluate and save {query} to {path}")
            with open(path, "wb") as f:
                f.write(b)

        if target_resource_directory is not None:
            filename = os.path.split(path)[1]
            key = (
                filename
                if target_resource_directory == ""
                else target_resource_directory + "/" + filename
            )
            print(f"*** Store evaluated {query} to {key}")
            if store is None:
                store = self.store()

            store.store(key, b, state.metadata)

        return state

    def evaluate_template(self, template: str, prefix="$", sufix="$", cache=None):
        """Evaluate a string template; replace all queries by their values
        Queries in the template are delimited by prefix and sufix.
        Queries should evaluate to strings and should not cause errors.
        """
        local_cache = {}
        result = ""
        for text, q in find_queries_in_template(template, prefix, sufix):
            result += text
            if q is not None:
                if q in local_cache:
                    result += local_cache[q]
                else:
                    state = self.evaluate(q, description=f"template expansion of {q}")
                    if state.is_error:
                        self.error(f"Template failed to expand {q}")
                        qr = f"ERROR({q})"
                    else:
                        qr = str(state.get())
                    local_cache[q] = qr
                    result += qr
        return result




from urllib.parse import quote, unquote
import re
from pyparsing import (
    Literal,
    Word,
    alphas,
    alphanums,
    nums,
    delimitedList,
    Regex,
    OneOrMore,
    Group,
    Combine,
    ZeroOrMore,
    Optional,
    Forward,
    lineno,
    col,
    FollowedBy,
)

"""
This module provides functions to decode and encode query.
A query is a sequence of commands, each command being a list of strings starting with the command name as a first element.
Commands are separated by a COMMAND_SEPARATOR, parts of the command (tokens) by PARAMETER_SEPARATOR.
Having special charachers (like COMMAND_SEPARATOR or PARAMETER_SEPARATOR) inside the commands is possible
by escaping with ESCAPE character and using escape sequences.
Query is URL-safe and URL escaping is applied to other special characters.
Empty commands are ignored (e.g. two consecutive COMMAND_SEPARATORs are equivalent to a single COMMAND_SEPARATOR).

Query can be decoded from string to a list of lists of strings by decode (or encoded by encode).
Encode and decode use encode_token and decode_token that are acting on a token level.
"""

COMMAND_SEPARATOR = "/"
PARAMETER_SEPARATOR = "-"
ESCAPE = "~"
ESCAPE_SEQUENCES = [
    (ESCAPE, ESCAPE + ESCAPE),
    ("https://", ESCAPE + "H"),
    ("http://", ESCAPE + "h"),
    ("file://", ESCAPE + "f"),
    ("://", ESCAPE + "P"),
    (COMMAND_SEPARATOR, ESCAPE + "I"),
    (PARAMETER_SEPARATOR, ESCAPE + "_"),
    (" ", ESCAPE + "."),
]


def decode(query: str):
    """Decode query string into a list of lists of strings."""
    ql = [
        [decode_token(etoken) for etoken in eqv.split(PARAMETER_SEPARATOR)]
        for eqv in query.split(COMMAND_SEPARATOR)
    ]
    return [qcommand for qcommand in ql if len(qcommand) and len(qcommand[0])]


def encode_token(token: str):
    "Encode single token by escaping special characters and character sequences"
    for sequence, encoding in ESCAPE_SEQUENCES:
        token = token.replace(sequence, encoding)
    return quote(token).replace("%7E", "~").replace("%7e", "~")


def decode_token(token: str):
    "Decode single token by un-escaping special characters and character sequences"
    encoding = {e: s for s, e in ESCAPE_SEQUENCES}
    if token == "":
        return ""
    try:
        index = token.index(ESCAPE)
        head = token[:index]
        mid = token[index : index + 2]
        tail = token[index + 2 :]
        return unquote(head + encoding.get(mid, mid)) + decode_token(tail)
    except ValueError:
        return unquote(token)


def encode(ql: list):
    """Decode query list (list of lists of strings) into an properly escaped query string."""
    return COMMAND_SEPARATOR.join(
        PARAMETER_SEPARATOR.join(encode_token(token) for token in qv) for qv in ql
    )


def all_splits(query):
    """Make all splits of the query into two parts (biginning and remainder)
    by progressively increasing the remainder.
    query == beginning + COMMAND_SEPARATOR + remainder
    """
    ql = decode(query)
    for i in range(len(ql), -1, -1):
        yield encode(ql[:i]), encode(ql[i:])


def indent(text, prefix="  "):
    return "\n".join(prefix + x for x in text.split("\n"))


def list_indent(lst, prefix="  "):
    if len(lst):
        return "[\n" + ",\n".join(indent(repr(x), prefix) for x in lst) + "\n]"
    else:
        return "[]"


class Position:
    def __init__(self, offset=0, line=0, column=0):
        self.offset = offset
        self.line = line
        self.column = column

    @classmethod
    def from_loc(cls, loc, string):
        return cls(loc, line=lineno(loc, string), column=col(loc, string))

    @classmethod
    def from_dict(cls, d):
        if d is None:
            return cls()
        else:
            return cls(offset=d["offset"], line=d["line"], column=d["column"])

    def to_dict(self):
        return dict(offset=self.offset, line=self.line, column=self.column)

    def __str__(self):
        if self.line == 0:
            return "(unknown position)"
        elif self.line > 1:
            return f"line {self.line}, position {self.column}"
        else:
            return f"position {self.column}"

    def __repr__(self):
        if self.line == 0:
            return "Position()"
        return f"Position(offset={self.offset}, line={self.line}, column={self.column})"


class QueryException(Exception):
    def __init__(self, message, position=None, query=None):
        self.original_message = message
        if position is not None:
            message += f" at {position}"
            if query is not None:
                message += f":\n  query: {query[:position.offset]}\n     {' '*position.offset}--> {query[position.offset:]}"
        else:
            if query is not None:
                message += f":\n  query: {query}"

        super().__init__(message)
        self.position = position
        self.query = query


class ActionParameter(object):
    def __init__(self, position=None):
        self.position = position or Position()

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.position)})"

    def __str__(self):
        return f"{self.__class__.__name__} at {self.position}"


class LinkActionParameter(ActionParameter):
    def __init__(self, link, position=None):
        super().__init__(position)
        self.link = link

    def encode(self):
        return "~X~" + self.link.encode() + "~E"

    def __repr__(self):
        return f"""LinkActionParameter(
{indent(repr(self.link))},
  {repr(self.position)})"""

    def __str__(self):
        return self.encode()


class ExpandedActionParameter(ActionParameter):
    def __init__(self, value, link, position=None):
        super().__init__(position)
        self.value = value
        self.link = link

    def encode(self):
        return "~X~" + self.link.encode() + "~E"

    def __repr__(self):
        return f"""ExpandedActionParameter(
{indent(repr(self.value))}
{indent(repr(self.link))},
  {repr(self.position)})"""

    def __str__(self):
        return str(self.value)


class StringActionParameter(ActionParameter):
    def __init__(self, string: str, position=None):
        super().__init__(position)
        self.string = string

    def encode(self):
        #        encoded = self.string.replace("~","~~").replace("/","~/").replace("://","~P")
        encoded = encode_token(self.string)
        return encoded

    def __repr__(self):
        return f"StringActionParameter({repr(self.string)}, {repr(self.position)})"

    def __str__(self):
        return self.string


class ResourceName(ActionParameter):
    def __init__(self, name: str, position=None):
        self.position = position or Position()
        self.name = name

    def encode(self):
        return self.name

    def __repr__(self):
        return f"ResourceName({repr(self.name)}, {repr(self.position)})"

    def __str__(self):
        return self.encode()


class ActionRequest(object):
    def __init__(self, name: str, parameters=None, position=None):
        self.name = name
        self.parameters = [] if parameters is None else parameters
        self.position = position or Position()

    @classmethod
    def from_arguments(cls, name: str, *parameters):
        assert type(name) == str
        typedparam = []
        for p in parameters:
            if isinstance(p, ActionParameter):
                typedparam.append(p)
            else:
                assert type(p) in (str, int, float, bool)
                typedparam.append(StringActionParameter(str(p)))

        return cls(name, typedparam)

    def to_list(self):
        lst = [self.name]
        for x in self.parameters:
            if isinstance(x, StringActionParameter):
                lst.append(x.string)
            elif isinstance(x, LinkActionParameter):
                lst.append(x.encode())
            else:
                raise Exception(
                    f"Unsupported action parameter type: {type(x)} ({repr(x)})"
                )
        return lst

    @classmethod
    def from_list(cls, lst):
        return cls.from_arguments(lst[0], *lst[1:])

    def encode(self):
        if len(self.parameters):
            p = "-".join(x.encode() for x in self.parameters)
            return f"{self.name}-{p}"
        else:
            return self.name

    def __repr__(self):
        return f"""ActionRequest(
  {repr(self.name)},
{indent(list_indent(self.parameters))},
  {repr(self.position)}
)"""

    def __str__(self):
        return self.encode()


class SegmentHeader(object):
    """Header of a query segment - both resource and transformation query.
    Header may contain name (string), level (integer) and parameters (list of strings).
    The header parameters may influence how the query is interpreted.
    The interpretation of the header parameters depends on the context object.
    """
    def __init__(
        self,
        name: str = "",
        level: int = 1,
        parameters: list = None,
        resource: bool = False,
        position=None,
    ):
        self.name = name
        self.level = level
        self.parameters = parameters or []
        self.resource = resource
        self.position = position or Position()

    def is_trivial(self):
        """Terurns true if the header does not contain any data,
        I.e. trivial header has no name, level is 1 and no parameters.
        Trivial header can be both for resource and query, it does not depend on the resource flas. 
        """
        return self.name in ("", None) and self.level==1 and len(self.parameters)==0

    def encode(self):
        assert self.level >= 1
        encoded = "-" * self.level
        if self.resource:
            encoded += "R"
        encoded += self.name
        if len(self.parameters):
            assert len(self.name) > 0 or self.resource
            for parameter in self.parameters:
                encoded += "-"
                encoded += parameter.encode()
        return encoded

    def __repr__(self):
        return f"""SegmentHeader(
  name       ={repr(self.name)},
  level      ={self.level},
  parameters ={indent(list_indent(self.parameters))},
  resource   ={self.resource},
  position   ={repr(self.position)}
)"""

    def __str__(self):
        return self.encode()


class TransformQuerySegment(object):
    """Query segment representing a transformation, i.e. a sequence of actions applied to a state."""
    def __init__(self, header=None, query=None, filename=None):
        "header can be SegmentHeader, query is a list of ActionRequest objects"
        self.header = header
        self.query = query or []
        self.filename = filename

    def predecessor(self):
        if self.filename is None:
            if len(self.query):
                p = TransformQuerySegment(
                    header=self.header, query=self.query[:-1], filename=None
                )
                r = TransformQuerySegment(
                    header=self.header, query=[self.query[-1]], filename=None
                )
                return p, r
            else:
                return None, None
        else:
            p = TransformQuerySegment(
                header=self.header, query=self.query, filename=None
            )
            r = TransformQuerySegment(
                header=self.header, query=[], filename=self.filename
            )
            return p, r

    def is_empty(self):
        return len(self.query) == 0 and self.filename is None

    def is_filename(self):
        return len(self.query) == 0 and self.filename is not None

    def is_action_request(self):
        return len(self.query) == 1 and self.filename is None

    def action(self):
        if self.is_action_request():
            return self.query[0]
        else:
            return None

    def encode(self):
        query = "/".join(x.encode() for x in self.query)
        if self.filename is not None:
            query = f"{query}/{self.filename}" if len(query) else self.filename

        if self.header is None:
            return query
        else:
            if len(query):
                return f"{self.header.encode()}/{query}"
            else:
                return self.header.encode()

    def append(self, q):
        if q is None:
            return self
        if isinstance(q, TransformQuerySegment):
            self.query.extend(q.query)
            self.filename = q.filename
            return self
        if isinstance(q, Query):
            if q.is_transform_query():
                return self.append(q.segments[0])
            else:
                raise Exception(
                    f"Appending general query {q} to transform query {self.encode()} not supported"
                )
        if isinstance(q, ActionRequest):
            self.query.append(q)
            return self
        if isinstance(q, str):
            return self.append(parse(q))

        raise Exception(
            f"Transform query {self.encode()} can't append object {repr(q)}"
        )

    def __add__(self, q):
        if q is None:
            return self
        if isinstance(q, TransformQuerySegment):
            return TransformQuerySegment(
                header=self.header, query=self.query + q.query, filename=q.filename
            )

        raise Exception(f"Unsupported operation (add): {self.encode()} + {repr(q)}")

    def __radd__(self, q):
        if q is None:
            return self
        if isinstance(q, TransformQuerySegment):
            return q + self
        raise Exception(f"Unsupported operation (radd): {repr(q)} + {self.encode()}")

    def __repr__(self):
        return f"""TransformQuerySegment(
  header   = {indent(repr(self.header))},
  query    = {indent(list_indent(self.query))},
  filename = {repr(self.filename)}
)"""

    def __str__(self):
        return self.encode()


class ResourceQuerySegment(object):
    """Query segment representing a resource, i.e. path to a file in a store."""
    def __init__(self, header=None, query=None):
        "header can be SegmentHeader, query is a list of ActionRequest objects"
        self.header = header
        self.query = query or []

    @property
    def position(self):
        if self.header is not None:
            return self.header.position
        else:
            if len(self.query):
                return self.query[0].position

    def path(self):
        """Path to the resource as a string.
        This is typically interpreted as a resource key in a Store object."""
        return "/".join(x.encode() for x in self.query)

    def encode(self):
        query = self.path()
        if self.header is None:
            rqs = ""
        else:
            rqs = self.header.encode()
        if len(rqs):
            rqs += "/"
        if len(query):
            return f"{rqs}{query}"
        else:
            return rqs

    def __repr__(self):
        return f"""ResourceQuerySegment(
  header = {indent(repr(self.header))},
  query  = {indent(list_indent(self.query))}
)"""

    def __str__(self):
        return self.encode()


class Query(object):
    """Query is a sequence of query segment.
    Typically this will be a resource and and/or a transformation applied to a resource."""
    def __init__(self, segments: list = None, absolute=False):
        self.segments = segments or []
        self.absolute = absolute

    def filename(self):
        if len(self.segments):
            segment = self.segments[-1]
            if (
                isinstance(segment, TransformQuerySegment)
                and segment.filename is not None
            ):
                return str(segment.filename)
            if (
                isinstance(segment, ResourceQuerySegment)
                and segment.query is not None
                and len(segment.query)
            ):
                return str(segment.query[-1])
        return None

    def extension(self):
        filename = self.filename()
        if filename is not None:
            v = filename.split(".")
            if len(v)>1:
                return v[-1]
        return None
    def is_empty(self):
        return len(self.segments) == 0

    def is_transform_query(self):
        return len(self.segments) == 1 and isinstance(
            self.segments[0], TransformQuerySegment
        )

    def transform_query(self):
        if self.is_transform_query():
            return self.segments[0]
        else:
            return None

    def is_resource_query(self):
        return len(self.segments) == 1 and isinstance(
            self.segments[0], ResourceQuerySegment
        )

    def resource_query(self):
        if self.is_resource_query():
            return self.segments[0]
        else:
            return None

    def is_action_request(self):
        return self.is_transform_query() and self.segments[0].is_action_request()

    def action(self):
        if self.is_action_request():
            return self.segments[0].action()
        else:
            return None

    def predecessor(self):
        if len(self.segments):
            if isinstance(self.segments[-1], TransformQuerySegment):
                p, r = self.segments[-1].predecessor()
                if p is None or p.is_empty():
                    qp = Query(self.segments[:-1], absolute=self.absolute)
                    qr = r
                    return qp, qr
                else:
                    qp = Query(self.segments[:-1] + [p], absolute=self.absolute)
                    qr = r
                    return qp, qr
            else:
                return None, None
        else:
            return None, None

    def short(self):
        _, r = self.predecessor()
        if r is None:
            q = str(self)
            if len(self) > 30:
                q = "..." + q[-30:]
            return q
        else:
            return str(r)

    def all_predecessors(self):
        qp, qr = self, None
        while qp is not None:
            yield qp, qr
            qp, r = qp.predecessor()
            qr = r + qr

    def create_segment(self, name: str = "", level=1):
        qs = TransformQuerySegment(SegmentHeader(name, level=level))
        self.segments.append(qs)
        return qs

    def last_transform_query_segment(self):
        if not self.is_empty():
            if isinstance(self.segments[-1], TransformQuerySegment):
                return self.segments[-1]
        return self.create_segment()

    def with_action(self, name: str, *parameters):
        self.last_transform_query_segment().append(
            ActionRequest.from_arguments(name, *parameters)
        )
        return self

    def __add__(self, tq):
        assert isinstance(tq, TransformQuerySegment)
        return Query(self.segments + [tq], absolute=self.absolute)

    def encode(self):
        q=("/" if self.absolute else "") + "/".join(
            x.encode() for x in self.segments
        )
        if self.is_resource_query():
            if not q.startswith("-"):
                q="-R/"+q
        return q

    def __repr__(self):
        return f"""Query(
{indent(list_indent(self.segments))},
  absolute = {self.absolute}
)"""

    def __str__(self):
        return self.encode()


identifier = Regex("[a-z_][a-zA-Z0-9_]*").setName("identifier")
filename = Regex(r"[a-zA-Z0-9_]*\.[a-zA-Z0-9._-]*").setName("filename")


def _resource_name_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    name = unquote("".join(toks))
    return ResourceName(name, position=position)


resource_name = (
    Regex(r"[a-zA-Z0-9._][a-zA-Z0-9._-]*")
    .setParseAction(_resource_name_parse_action)
    .setName("resource_name")
)
parameter_text = Regex("[a-zA-Z0-9_+.]+").setName("parameter_text")
percent_encoding = Regex("%[0-9a-fA-F][0-9a-fA-F]").setName("percent_encoding")
parse_query = Forward()

tilde_entity = (
    Literal("~~").setParseAction(lambda s, loc, toks: ["~"]).setName("tilde_entity")
)
minus_entity = (
    Literal("~_").setParseAction(lambda s, loc, toks: ["-"]).setName("minus_entity")
)
islash_entity = (
    Literal("~I").setParseAction(lambda s, loc, toks: ["/"]).setName("islash_entity")
)
slash_entity = (
    Literal("~/").setParseAction(lambda s, loc, toks: ["/"]).setName("slash_entity")
)

https_entity = (
    Literal("~H")
    .setParseAction(lambda s, loc, toks: ["https://"])
    .setName("https_entity")
)
http_entity = (
    Literal("~h")
    .setParseAction(lambda s, loc, toks: ["http://"])
    .setName("http_entity")
)
file_entity = (
    Literal("~f")
    .setParseAction(lambda s, loc, toks: ["file://"])
    .setName("file_entity")
)
protocol_entity = (
    Literal("~P")
    .setParseAction(lambda s, loc, toks: ["://"])
    .setName("protocol_entity")
)

negative_number_entity = (
    Regex("~[0-9]")
    .setParseAction(lambda s, loc, toks: ["-" + toks[0][1:]])
    .setName("negative_number_entity")
)
space_entity = (
    Literal("~.").setParseAction(lambda s, loc, toks: [" "]).setName("space_entity")
)
end_entity = Literal("~E")


def _expand_entity_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    return LinkActionParameter(toks[0], position=position)


expand_entity = (
    (Literal("~X~").suppress() + parse_query + end_entity.suppress())
    .setParseAction(_expand_entity_parse_action)
    .setName("expand_entity")
)

entities = (
    tilde_entity
    | minus_entity
    | negative_number_entity
    | space_entity
    | islash_entity
    | slash_entity
    | http_entity
    | https_entity
    | file_entity
    | protocol_entity
)


def _parameter_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    par = unquote("".join(toks))
    return StringActionParameter(par, position=position)


parameter = expand_entity | (
    ZeroOrMore(parameter_text | entities | percent_encoding)
    .setParseAction(_parameter_parse_action)
    .setName("parameter")
)


def _action_request_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    name = toks[0]
    parameters = list(toks[1:])
    return ActionRequest(name=name, parameters=parameters, position=position)


action_request = (
    (identifier + ZeroOrMore(Literal("-").suppress() + parameter))
    .setParseAction(_action_request_parse_action)
    .setName("action_request")
)


def _action_path_parse_action(s, loc, toks):
    ap = list(toks)
    if type(ap[-1]) is str:
        return (ap[:-1], ap[-1])
    else:
        return (ap, None)


# segment_indicator = (
#    OneOrMore(Literal("-"))
#    .setParseAction(lambda s, loc, toks: len(toks))
#    .setName("segment_indicator")
# )


def _segment_identifier_action(s, loc, toks):
    m = re.match("(-+)([a-zA-Z0-9_]*)", toks[0])
    assert m is not None
    return len(m.group(1)), m.group(2)


segment_identifier = (
    (Regex("-+[a-z][a-zA-Z0-9_]*") | (Regex("-+") + FollowedBy("/")))
    .setParseAction(_segment_identifier_action)
    .setName("segment_identifier")
)


def _resource_identifier_action(s, loc, toks):
    m = re.match("(-+)R([a-zA-Z0-9_]*)", toks[0])
    assert m is not None
    return len(m.group(1)), m.group(2)


resource_identifier = (
    (Regex("-+R[a-zA-Z0-9_]*"))
    .setParseAction(_resource_identifier_action)
    .setName("resource_identifier")
)

action_path_nonempty = (
    (
        (
            ZeroOrMore(
                action_request
                + (
                    Literal("/") + ~(resource_identifier | segment_identifier)
                ).suppress()
            )
            + (filename | action_request)
        )
    )
    .setParseAction(_action_path_parse_action)
    .setName("action_path_nonempty")
)


def _resource_path_parse_action(s, loc, toks):
    return list(toks)


resource_path = (
    delimitedList(resource_name, delim="/")
    .setParseAction(_resource_path_parse_action)
    .setName("resource_path")
)


def _segment_header_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    level, name = toks[0]
    if len(toks) > 1:
        parameters = list(toks[1:])
        return SegmentHeader(
            name=name, level=level, parameters=parameters, position=position
        )
    else:
        return SegmentHeader(level=level, name=name)


segment_header = (
    (segment_identifier + ZeroOrMore(Literal("-").suppress() + parameter))
    .setParseAction(_segment_header_parse_action)
    .setName("segment_header")
)


def _resource_segment_with_header_parse_action(s, loc, toks):
    position = Position.from_loc(loc, s)
    level, name = toks[0]
    header = SegmentHeader(
        name=name,
        level=level,
        parameters=list(toks[1]),
        resource=True,
        position=position,
    )
    if len(toks) == 3:
        return ResourceQuerySegment(header=header, query=list(toks[2]))
    else:
        return ResourceQuerySegment(header=header, query=[])


resource_segment_with_header = (
    (
        resource_identifier
        + Group(ZeroOrMore(Word("-").suppress() + parameter))
        + Optional(Literal("/").suppress()+Group(resource_path))
    )
    .setParseAction(_resource_segment_with_header_parse_action)
    .setName("resource_segment_with_header")
)


def _segment_with_header_parse_action(s, loc, toks):
    if len(toks) == 1:
        return TransformQuerySegment(header=toks[0])
    else:
        assert len(toks) == 3
        return TransformQuerySegment(
            header=toks[0], query=toks[2][0][0], filename=toks[2][0][1]
        )


segment_with_header = (
    (segment_header + Optional(Literal("/") + Group(action_path_nonempty)))
    .setParseAction(_segment_with_header_parse_action)
    .setName("segment_with_header")
)


def _segment_without_header_parse_action(s, loc, toks):
    return TransformQuerySegment(query=list(toks[0][0][0]), filename=toks[0][0][1])


segment_without_header = (
    Group(action_path_nonempty)
    .setParseAction(_segment_without_header_parse_action)
    .setName("segment_without_header")
)

query_segment = (
    segment_with_header | segment_without_header | resource_segment_with_header
).setName("query_segment")


def _parse_query_parse_action(s, loc, toks):
    if toks[0] == "/":
        return Query(segments=list(toks[1:]), absolute=True)
    else:
        return Query(segments=list(toks), absolute=False)


parse_query << (
    (Optional(Literal("/")) + delimitedList(query_segment, "/"))
    .setParseAction(_parse_query_parse_action)
    .setName("parse_query")
)


def _resource_transform_query_action(s, loc, toks):
    if toks[0] == "/":
        assert len(toks) == 3
        resource = ResourceQuerySegment(query=list(toks[1]))
        transform = toks[2]
        return Query(segments=[resource, transform], absolute=True)
    else:
        assert len(toks) == 2
        resource = ResourceQuerySegment(query=list(toks[0]))
        transform = toks[1]
        return Query(segments=[resource, transform], absolute=False)


resource_transform_query = (
    (
        Optional(Literal("/"))
        + Group(resource_path)
        + Literal("/").suppress()
        + segment_with_header
    )
    .setParseAction(_resource_transform_query_action)
    .setName("resource_transform_query")
)


def parse(query):
    try:
        return resource_transform_query.parseString(query, True)[0]
    except:
        return parse_query.parseString(query, True)[0]


if __name__ == "__main__":
    #    print(action_path_nonempty.parseString("abc/def/file.txt", True))
    #    print(query_segment.parseString("abc/def/file.txt", True))
    # print(resource_path.parseString("abc/def/file.txt", True))
    #    print(query_segment.parseString("-qs/abc/def/file.txt", True))
    #    print(parse_query.parseString("abc/def/file.txt", True))
    #    print(parse_query.parseString("-/abc/def/file.txt", True))
    #    print(parse_query.parseString("-qs/abc/def/file.txt", True))
    #    print(parse_query.parseString("-/xxx", True))
    #    print(parse_query.parseString("abc/def/-/xxx", True))
    #    print(parse_query.parseString("abc/def/-/xxx/file.txt", True))
    #    print(repr(parse("abc/def/-/xxx/file.txt")))
    #    print(repr(parse("abc/def")))

    #    print(parse("abc-def/-/x-y/--xxx-y/aaa"))
    #    print(repr(parse("/abc-def/-/x-y/--xxx-y/aaa")))
    #    print(resource_segment_with_header.parseString("-R/abc/def"))
    # print(resource_segment_with_header.parseString("-R-1-2/"))
    print(repr(parse("-R/abc/def/-/ghi")))
    print((parse("-R/abc/def/-/ghi")))
    print()
    for p, r in parse("-R/abc/def/-/ghi/jkl/file.txt").all_predecessors():
        print(p.encode(), "   -   ", "NONE" if r is None else r.encode())
    #    print(parse("abc/xxx/-/def"))
    #    print(repr(parse("-/def")))
    #    print(repr(parse("def")))
    #    print(repr(action_path_nonempty.parseString("abc", True)))
    #    print(repr(expand_entity.parseString("~X~abc~E", True)))
    #    print(repr(parse("-/def-~X~abc~E")))
    print(parse("abc/def/-/xxx/-q/qqq"))
    print(parse("abc/def/-/xxx/-q/qqq-abc-~X~xxx/yyy~E-def"))
    print(repr(parse("abc/def/-/xxx/-q/qqq-abc-~X~xxx/yyy~E-def")))



import traceback
import os.path
from liquer.context import get_context, find_queries_in_template


def evaluate(query):
    """Evaluate query, returns a State, cache the output in supplied cache"""
    return get_context().evaluate(query)


def evaluate_and_save(
    query, target_directory=None, target_file=None, target_resource_directory=None
):
    """Evaluate query and save result.
    Output is saved either to
    - a target directory (current working directory by default) to a file deduced from the query, or
    - to target_file (if specified)
    Returns a state.
    """
    return get_context().evaluate_and_save(
        query,
        target_directory=target_directory,
        target_file=target_file,
        target_resource_directory=target_resource_directory,
    )


def evaluate_template(template: str, prefix="$", sufix="$"):
    """Evaluate a string template; replace all queries by their values
    Queries in the template are delimited by prefix and sufix.
    Queries should evaluate to strings and should not cause errors.
    """
    return get_context().evaluate_template(template, prefix=prefix, sufix=sufix)



import pandas as pd
import io
from urllib.request import urlopen
import numpy as np
from liquer.commands import command
import plotly.graph_objs as go
from plotly.offline import download_plotlyjs, plot


@command
def plotly_chart(state, *series):
    """Plotly chart"""
    series = list(reversed(list(series)))
    df = state.get()
    extension = "html"
    output_type = "page"
    include_plotlyjs = True
    html = ""

    while len(series):
        t = series.pop()
        if t in ["div"]:
            output_type = t
        elif t in ["cdn"]:
            include_plotlyjs = t
        elif t == "xy":
            xcol = series.pop()
            ycol = series.pop()
            state.log_info(f"Chart XY ({xcol} {ycol})")
            html = plot(
                [go.Scatter(x=df[xcol], y=df[ycol])], show_link=False, output_type="div"
            )
            continue
        elif t == "xys":
            xcol = series.pop()
            ycol = series.pop()
            scol = series.pop()
            state.log_info(f"Chart XY Set ({xcol} {ycol} {scol})")
            datasets = list(sorted(df[scol].unique()))
            p = []
            for ds in datasets:
                index = df[scol] == ds
                p.append(
                    go.Scatter(
                        x=df.loc[index, xcol],
                        y=df.loc[index, ycol],
                        mode="lines+markers",
                        name=ds,
                    )
                )
            html = plot(p, show_link=False, output_type="div")
        elif t == "piexs":
            xcol = series.pop()
            scol = series.pop()
            state.log_info(f"Pie X Set ({xcol} {scol})")
            html = plot(
                [
                    go.Pie(
                        labels=df[scol],
                        values=df[xcol],
                        hoverinfo="label+percent",
                        textinfo="value",
                    )
                ],
                show_link=False,
                output_type="div",
            )
        else:
            state.log_warning(f"Unrecognized plotly_chart parameter {t}")
    if output_type == "page":
        html = f"""<html><body>{html}</body></html>"""
    return state.with_data(html).with_filename(f"image.{extension}")



from liquer.commands import command, first_command, command_registry
import liquer.ext.basic
import pygments
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter


@command(ns="pygments")
def highlight(code, lexer="py", style="default"):
    """Syntax highlighting by Pygments
    Returns html-formatted code.
    Lexer specifies the syntax (currently only "py" for python).
    Style is the style parameter for the Pygments formatter.
    """
    if lexer == "py":
        html = pygments.highlight(
            code, PythonLexer(), HtmlFormatter(noclasses=True, style=style)
        )
    else:
        raise Exception(f"Unsupported lexer: {lexer}")
    return html



from liquer.commands import command, first_command, command_registry
from liquer.query import evaluate_template
import liquer.ext.basic
from liquer.context import *
import traceback


@first_command(ns="meta")
def commands():
    """Returns a list of commands in json format"""
    return command_registry().as_dict()


@first_command(ns="meta")
def flat_commands(*pairs):
    """Returns a list of commands in json format"""
    filters = [(pairs[i - 1], pairs[i]) for i in range(1, len(pairs), 2)]
    d = []
    for rl in command_registry().as_dict().values():
        for r in rl.values():
            if len(filters) == 0 or all(
                r["attributes"].get(key) == value for key, value in filters
            ):
                c = {key: r.get(key) for key in ["name", "label", "module", "doc"]}
                for key, value in r["attributes"].items():
                    if key not in c:
                        c[key] = value
                d.append(c)
    return d


@first_command(ns="meta")
def flat_commands_nodoc(*pairs):
    """Returns a list of commands in json format"""
    filters = [(pairs[i - 1], pairs[i]) for i in range(1, len(pairs), 2)]
    d = []
    for rl in command_registry().as_dict().values():
        for r in rl.values():
            if len(filters) == 0 or all(
                r["attributes"].get(key) == value for key, value in filters
            ):
                c = {key: r.get(key) for key in ["name", "label", "module"]}
                for key, value in r["attributes"].items():
                    if key not in c and key != "context_menu":
                        c[key] = value
                d.append(c)
    return d


@command
def state(state):
    """Returns dictionary with all metadata for the resulting state"""
    return state.as_dict()

@command(ns="meta")
def metadata(state):
    """Returns dictionary with all metadata for the resulting state"""
    return {**state.metadata}

@command(ns="meta")
def help(state, command_name, ns="root"):
    """Returns a description of the command"""
    crd = command_registry().as_dict()
    make_link = ("py" in crd) and ("pygments" in crd)
    try:
        r = crd[ns]
        try:
            c = r[command_name]
            html = f"<html><head><title>{command_name}</title></head>"
            html = "<body\n>"
            html = f"<h1>{command_name} <em>({ns})</em></h1>\n"
            html += f"<pre>{c['doc']}</pre>\n"
            html += "<h3>Arguments</h3>\n"
            html += "  <ul>\n"
            for a in c["arguments"]:
                html += f"    <li><b>{a['name']}</b> - type: {a['type']}</li>\n"
            html += "  </ul>"
            html += "<h3>Attributes</h3>\n"
            html += "  <ul>\n"
            for key, value in sorted(c["attributes"].items()):
                if key == "example_link":
                    q = value
                    url = (
                        state.vars.get("server", "http://localhost")
                        + state.vars.get("api_path", "/q/")
                        + q
                    )
                    html += f'    <li><b>{key}</b>:<a href="{url}">{value}</a></li>\n'
                else:
                    html += f"    <li><b>{key}</b>:{value}</li>\n"
            html += "  </ul>"
            html += f"<h3>Python module</h3>\n"
            module = c["module"]
            if make_link:
                modfile = module.replace(".", "_")
                html += evaluate_template(
                    f'<a href="$ns-py-pygments/module_source-{module}/highlight/link-url-html$/{modfile}.html">{module}</a>\n'
                )
            else:
                html += module

            html += "</body>"
            html += "</html>"
            return state.with_data(html).with_filename("help.html")
        except KeyError:
            html = f"<h1>Command <em>{command_name}</em> not registered in namespace <em>{ns}</em></h1>"
            candidates = ", ".join(
                key for key, value in crd.items() if command_name in value
            )
            if len(candidates):
                html += f"Command can be found in: {candidates}"
            else:
                html += "There is no such command in available namespaces"
            return state.with_data(html).with_filename("help.html")
    except KeyError:
        return state.with_data(
            f"<h1>Unknown namespace <em>{ns}</em></h1>"
        ).with_filename("help.html")

@command(ns="meta")
def status_md(metadata):
    txt = ""
    txt+="# %s%s\n\n"%(metadata.get("title","???"), " (ERROR)" if metadata.get("is_error") else "")
    if metadata.get("key") is not None:
        txt+="KEY:             %s\n"%(metadata.get("key",""))
        if metadata.get("has_recipe",False):
            txt+="Resource has a recipe\n"

    txt+="QUERY:           %s\n"%(metadata.get("query","???"))
    txt+="STATUS:          %s\n"%(metadata.get("status","???"))
    txt+="TYPE IDENTIFIER: %s\n"%(metadata.get("type_identifier","???"))
    txt+="MIME:            %s\n"%(metadata.get("mimetype","???"))
    txt+="PARENT:          %s\n"%(metadata.get("parent_query","???"))
    txt+="\nMESSAGE:         %s\n\n"%(metadata.get("message",""))

    txt+="STARTED: %s\n"%(metadata.get("started","-"))
    txt+="UPDATED: %s\n"%(metadata.get("updated","-"))
    txt+="CREATED: %s\n"%(metadata.get("created","-"))
    txt+="\n"

    fileinfo = metadata.get("fileinfo")
    if fileinfo is not None:
        txt+="\n## FILEINFO%s\n"%(" (DIRECTORY)" if fileinfo.get("is_dir") else "")
        txt+="NAME   : %s\n"%(fileinfo.get("name",""))
        txt+="PATH   : %s\n"%(fileinfo.get("filesystem_path","-"))
        txt+="SIZE   : %s\n"%(fileinfo.get("size","?"))
        txt+="\n"

    txt+="\n## DESCRIPTION:\n%s\n"%(metadata.get("description",""))

    
    txt+="\n## DATA CHARACTERISTICS:\n%s\n\n"%(metadata.get("data_characteristics",{}).get("description",""))

    for key,value in sorted(metadata.get("data_characteristics",{}).items()):
        if key not in ("description","type_identifier"):
            txt+="%-20s:%s\n"%(key, repr(value))

    txt+="\n## STATE VARIABLES\n"
    for key, value in sorted(metadata.get("vars",{}).items()):
        txt+="%-20s:%s\n"%(key, repr(value))
    
    txt+="\n## LOGS\n"
    for name, log in [
        ("Main log",metadata.get("log")),
        ("Child log",metadata.get("child_log")),
        ("Resource log",metadata.get("resource_metadata",{}).get("log")),
        ("Resource child log",metadata.get("resource_metadata",{}).get("child_log"))
        ]:
        if log is not None and len(log):
            txt+=f"\n### {name}\n"
            for entry in log:
                if type(entry)!=dict:
                    txt+=f"INVALID ENTRY {repr(entry)}\n"
                else:
                    txt+="%-10s %-28s %s\n"%(entry.get("kind","?????"), entry.get("timestamp",""), entry.get("origin",""))
                    txt+="%-10s %s\n"%("",entry.get("message"))
                    tb = entry.get('traceback')
                    if tb is not None:
                        if len(tb):
                            txt+=tb
                            txt+="\n"

#    txt+="## LOG\n"
#    for record in metadata.get("log",[]):
#        txt+=str(record)
    return txt

@command(ns="meta", volatile=True)
def dir_status(metadata, context=None):
    context = get_context(context)
    if metadata is None:
        key=""
    else:
        key = metadata.get("key")
    print("context", context)
    context.set_title(f"Status of directory {key}")
    context.set_description(f"Status overview of data in the directory {key}")
    store = context.store()
    data = []
    if store.is_dir(key):
        for name in store.listdir(key):
            metadata = store.get_metadata(store.join_key(key,name))
            fileinfo = metadata.get("fileinfo",{})
            data.append(dict(
                name=name,
                key=metadata.get("key"),
                query=metadata.get("query"),
                title=metadata.get("title"),
                description=metadata.get("description"),
                status=metadata.get("status"),
                is_error=metadata.get("is_error"),
                message=metadata.get("message",""),
                data_characteristics=metadata.get("data_characteristics",{}).get("description",""),
                is_dir=fileinfo.get("is_dir"),
                size=fileinfo.get("size"),
                started=metadata.get("started"),
                created=metadata.get("created"),
                updated=metadata.get("updated"),
                has_recipe=metadata.get("has_recipe",False),
                ))
        return dict(status="OK", message="OK", data=data)
    else:
        return dict(status="ERROR", message=f"Not a directory:{key}", data=[])

@command(ns="meta")
def dir_status_df(metadata, context=None):
    import pandas as pd
    context = get_context(context)
    data = dir_status(metadata, context=context)
    return pd.DataFrame(data["data"], columns=[
        "name", "title", "status", "is_error", "message", "data_characteristics", "started", "created", "updated",
        "key", "query", "description", "is_dir", "size"])

@command(ns="meta", volatile=True)
def clean_recipes(metadata, recursive=False, context=None):
    """Remove specific key or all the data in the directory of store that has a recipe.
    This is supposed to be used together with a RecipeStore, that creates has_recipe flag in the metadata.
    Data can be optionally cleaned recursively in all the subdirectories.

    Examples:
    -R-meta/my_dir/-/ns-meta/clean_recipes    - deletes all data with recipes in my_dir (but not its subfolders)
    -R-meta/my_dir/-/ns-meta/clean_recipes-t  - deletes all data with recipes in my_dir and its subfolders
    -R-meta/-/ns-meta/clean_recipes-t         - clean all the data with recipes everywhere
    """
    context = get_context(context)

    if not isinstance(metadata, dict) or "key" not in metadata:
        raise Exception("Valid metadata with a key expected in clean_recipes")

    key = metadata["key"]
    store=context.store()
    removed=[]
    if store.is_dir(key):
        if key in ("",None):
            if recursive:
                keys = store.keys()
            else:
                keys = store.listdir(key)
        else:
            if recursive:
                keys = [k for k in store.keys() if k.startswith(f"{key}/")]
            else:
                keys = [f"{key}/{name}" for name in store.listdir(key)]
    for key in keys:
        if store.is_dir(key):
            continue
        try:
            metadata = store.get_metadata(key)
            if metadata.get("has_recipe",False):
                context.info(f"Remove {key}")
                store.remove(key)
                removed.append(key)
        except:
            context.warning(f"Failed to process {key} ")
            context.warning(traceback.format_exc())
    context.info(f"Removed {len(removed)} items")
    return dict(status="OK", message=f"Removed {len(removed)} items", removed=removed)

@command(ns="meta", volatile=True)
def make_recipes(metadata, recursive=False, context=None):
    """Make all the data in the directory of a store that has a recipe.
    This is supposed to be used together with a RecipeStore, that creates has_recipe flag in the metadata.
    Data can be optionally cleaned recursively in all the subdirectories.
    """
    context = get_context(context)

    if not isinstance(metadata, dict) or "key" not in metadata:
        raise Exception("Valid metadata with a key expected in clean_recipes")

    key = metadata["key"]
    store=context.store()
    processed=[]
    if store.is_dir(key):
        if key in ("",None):
            if recursive:
                keys = store.keys()
            else:
                keys = store.listdir(key)
        else:
            if recursive:
                keys = [k for k in store.keys() if k.startswith(f"{key}/")]
            else:
                keys = [f"{key}/{name}" for name in store.listdir(key)]
    for key in keys:
        if store.is_dir(key):
            continue
        try:
            metadata = store.get_metadata(key)
            if metadata.get("has_recipe",False) and metadata.get("status") in (Status.RECIPE.value, Status.ERROR.value, Status.EXPIRED.value):
                context.info(f"Make {key}")
                store.remove(key)
                store.get_bytes(key)
                processed.append(key)
        except:
            context.warning(f"Failed to process {key} ")
            context.warning(traceback.format_exc())
    context.info(f"Successfully processed {len(processed)} items")
    return dict(status="OK", message=f"Successfully processed {len(processed)} items", processed=processed)

@first_command(ns="meta")
def root_key():
    return dict(key="")

@command(ns="meta", volatile=True)
def clean(metadata, errors=True, expired=False, evaluation=False, recipes=False, recursive=False, context=None):
    """Remove specific key or all the data in the directory of store.
    This command can delete from the store items in the following states:
    - errors (default behaviour)
    - expired items
    - recipes (valid items with a recipe)
    """
    context = get_context(context)

    if not isinstance(metadata, dict) or "key" not in metadata:
        raise Exception("Valid metadata with a key expected in clean")

    key = metadata["key"]
    store=context.store()
    removed=[]
    if store.is_dir(key):
        if key in ("",None):
            if recursive:
                keys = store.keys()
            else:
                keys = store.listdir(key)
        else:
            if recursive:
                keys = [k for k in store.keys() if k.startswith(f"{key}/")]
            else:
                keys = list(store.listdir_keys(key))
    for key in keys:
        print(f"KEY {key}")
        if store.is_dir(key):
            continue
        try:
            metadata = store.get_metadata(key)
            status = metadata.get("status")
            if errors and status == Status.ERROR.value:
                store.remove(key)
                removed.append((key,status))
                continue
            if expired and status == Status.EXPIRED.value:
                store.remove(key)
                removed.append((key,status))
                continue
            if evaluation and status in (Status.EVALUATION.value, Status.EVALUATING_DEPENDENCIES.value, Status.EVALUATING_PARENT.value, Status.SUBMITTED.value):
                store.remove(key)
                removed.append((key,status))
                continue

            if recipes:
                if metadata.get("has_recipe",False):
                    context.info(f"Remove {key}")
                    store.remove(key)
                    removed.append((key,"recipe"))
        except:
            context.warning(f"Failed to process {key} ")
            context.warning(traceback.format_exc())
    context.info(f"Removed {len(removed)} items")
    return dict(status="OK", message=f"Removed {len(removed)} items", removed=removed)

if __name__ == "__main__":
    from liquer import *

    print(evaluate("ns-meta/flat_commands").get())
    # print(evaluate("ns-meta/commands/state").get()["commands"])
    # evaluate_and_save("ns-meta/commands/state/state.json")



import pptx
from io import BytesIO
import pandas as pd
from liquer.state_types import StateType, register_state_type
from liquer.constants import mimetype_from_extension

from liquer.commands import command, first_command
from tempfile import TemporaryDirectory
from pathlib import Path
from liquer.context import get_context
class ResilientBytesIO(BytesIO):
    "Workaround to prevent closing the stream"

    def close(self):
        pass  # Refuse to close to avoid pandas bug

    def really_close(self):
        super().close()

class PPTXPresentationStateType(StateType):
    def identifier(self):
        return "pptx_presentation"

    def default_extension(self):
        return "pptx"

    def is_type_of(self, data):
        if isinstance(data, pptx.presentation.Presentation):
            return True
        return False

    def as_bytes(self, data, extension=None):
        if extension is None:
            extension = self.default_extension()
        assert self.is_type_of(data)
        mimetype = mimetype_from_extension(extension)

        if extension == "pptx":
            output = ResilientBytesIO()
            data.save(output)
            b = output.getvalue()
            output.really_close()
            return b, mimetype
        else:
            raise Exception(
                f"Serialization: file extension {extension} is not supported by pptx Presentation type."
            )

    def from_bytes(self, b: bytes, extension=None):
        if extension is None:
            extension = self.default_extension()
        f = BytesIO()
        f.write(b)
        f.seek(0)
        if extension == "pptx":
            return pptx.Presentation(f)
        raise Exception(
            f"Deserialization: file extension {extension} is not supported by pptx Presentation type."
        )

    def copy(self, data):
        return self.from_bytes(self.as_bytes(data)[0])

    def data_characteristics(self, data):
        return dict(description=f"Presentation (pptx) with {len(data.slides)} slides.")
        

PPTX_PRESENTATION_STATE_TYPE = PPTXPresentationStateType()
register_state_type(pptx.Presentation, PPTX_PRESENTATION_STATE_TYPE)
register_state_type(pptx.presentation.Presentation, PPTX_PRESENTATION_STATE_TYPE)
